import datetime
from typing import Any, Optional, Callable

from datacode.models.pipeline.operations.analysis import AnalysisOptions
from datacode.models.source import DataSource

from datacode.models.pipeline.base import DataPipeline
from datacode.models.types import DataSourceOrPipeline


class DataAnalysisPipeline(DataPipeline):
    """
    A DataPipeline which starts from a single data source but does not produce a data source
    """

    def __init__(self, data_source: DataSourceOrPipeline, options: AnalysisOptions,
                 name: Optional[str] = None):
        super().__init__(
            [data_source],
            [options],
            name=name
        )

    @property
    def data_source(self) -> DataSource:
        return self.data_sources[0]

    @data_source.setter
    def data_source(self, source: DataSource):
        self.data_sources[0] = source



import pandas as pd
import pyexlatex.table as lt

from datacode.typing import IntSequenceOrNone, FloatSequenceOrNone
from datacode.summarize.subset.missing.summary.tables.fullsample.idcount import by_id_count_data_table
from datacode.summarize.subset.missing.summary.tables.fullsample.missingmore import missing_more_than_data_table
from datacode.summarize.subset.missing.summary.tables.fullsample.obs import obs_count_and_missing_data_table
from datacode.summarize.subset.missing.summary.tables.fullsample.periods import periods_data_table
from pyexlatex.table.models.spacing.columntable import ColumnPadTable

def missing_full_sample_summary_panel(df: pd.DataFrame, id_col: str, col_with_missings: str,
                                      missing_tolerances: IntSequenceOrNone = (0, 5, 10),
                                      missing_quantiles: FloatSequenceOrNone = None,
                                      missing_display_str: str = 'Missing', datevar: str = 'Date',
                                      pct_format_str: str = '.1f', period_display_name: str = 'Period') -> lt.Panel:

    obs_pct_dt = obs_count_and_missing_data_table(
        df,
        col_with_missings,
        missing_display_str=missing_display_str,
        pct_format_str=pct_format_str
    )

    by_id_count_dt = by_id_count_data_table(df, id_col)

    period_counts_dt = periods_data_table(
        df,
        id_col,
        datevar,
        period_display_name=period_display_name
    )

    missing_more_than_dt = missing_more_than_data_table(
        df,
        id_col,
        col_with_missings,
        missing_tolerances=missing_tolerances,
        missing_quantiles=missing_quantiles,
        missing_display_str=missing_display_str,
        period_display_name=period_display_name,
        pct_format_str=pct_format_str
    )

    # Create a single data table from the existing data tables
    combine_items = [
        obs_pct_dt,
        by_id_count_dt,
        period_counts_dt,
        missing_more_than_dt
    ]
    data_table = combine_items[0]
    for item in combine_items[1:]:
        data_table += ColumnPadTable()  # pad with a single column between each
        data_table += item

    panel = lt.Panel.from_data_tables(
        [data_table],
        name=f'{missing_display_str} {col_with_missings} Full Sample Summary'
    )

    return panel


from datacode.models.transform.specific.change import change_transform
from datacode.models.transform.specific.lag import lag_transform

DEFAULT_TRANSFORMS = [
    lag_transform,
    change_transform,
]



import pandas as pd
from pandas._testing import assert_frame_equal

import datacode.hooks as dc_hooks
from datacode import DataSource

from tests.pipeline.base import PipelineTest

COUNTER = 0


def increase_counter_hook(*args, **kwargs):
    global COUNTER
    COUNTER += 1
    return kwargs


def increase_counter_hook_discard_first_arg(*args):
    increase_counter_hook()
    return args[1:]


def increase_counter_hook_return_only_second_arg(*args):
    increase_counter_hook()
    return args[1]


def increment_df_a_variable_hook(source: DataSource, df: pd.DataFrame) -> pd.DataFrame:
    df['a'] += 1
    return df


class HooksTest(PipelineTest):

    def teardown_method(self, *args, **kwargs):
        super().teardown_method(*args, **kwargs)
        dc_hooks.reset_hooks()


class TestPipelineHooks(HooksTest):

    def test_no_pipeline_hook(self):
        counter_value = COUNTER
        dgp = self.create_generator_pipeline()
        dgp.execute()
        assert COUNTER == counter_value

    def test_reset_hook(self):
        counter_value = COUNTER
        dc_hooks.on_begin_execute_pipeline = increase_counter_hook
        dc_hooks.reset_hooks()
        dgp = self.create_generator_pipeline()
        dgp.execute()
        assert COUNTER == counter_value

    def test_on_begin_execute_pipeline(self):
        counter_value = COUNTER
        dc_hooks.on_begin_execute_pipeline = increase_counter_hook
        dgp = self.create_generator_pipeline()
        dgp.execute()
        assert COUNTER == counter_value + 1

    def test_on_end_execute_pipeline(self):
        counter_value = COUNTER
        dc_hooks.on_end_execute_pipeline = increase_counter_hook
        dgp = self.create_generator_pipeline()
        dgp.execute()
        assert COUNTER == counter_value + 1

    def test_on_begin_execute_operation(self):
        counter_value = COUNTER
        dc_hooks.on_begin_execute_operation = increase_counter_hook
        dgp = self.create_generator_pipeline()
        dgp.execute()
        assert COUNTER == counter_value + 1

    def test_on_end_execute_operation(self):
        counter_value = COUNTER
        dc_hooks.on_end_execute_operation = increase_counter_hook
        dgp = self.create_generator_pipeline()
        dgp.execute()
        assert COUNTER == counter_value + 1

    def test_on_begin_load_source(self):
        counter_value = COUNTER
        dc_hooks.on_begin_load_source = increase_counter_hook
        self.create_csv()
        ds = self.create_source(df=None)
        df = ds.df
        assert COUNTER == counter_value + 1

    def test_on_end_load_source(self):
        dc_hooks.on_end_load_source = increment_df_a_variable_hook
        self.create_csv()
        ds = self.create_source(df=None)
        df = ds.df
        assert_frame_equal(df, self.expect_df_no_rename_a_plus_1)

    def test_on_begin_apply_variable_transform(self):
        counter_value = COUNTER
        dc_hooks.on_begin_apply_variable_transform = increase_counter_hook_discard_first_arg
        self.create_csv()
        cols = self.create_columns(transform_data='series')
        ds = self.create_source(df=None, columns=cols)
        df = ds.df
        assert COUNTER == counter_value + 2

    def test_on_end_apply_variable_transform(self):
        counter_value = COUNTER
        dc_hooks.on_end_apply_variable_transform = increase_counter_hook_return_only_second_arg
        self.create_csv()
        cols = self.create_columns(transform_data='series')
        ds = self.create_source(df=None, columns=cols)
        df = ds.df
        assert COUNTER == counter_value + 2

    def test_on_begin_apply_source_transform(self):
        counter_value = COUNTER
        dc_hooks.on_begin_apply_source_transform = increase_counter_hook_return_only_second_arg
        dtp = self.create_transformation_pipeline()
        dtp.execute()
        assert COUNTER == counter_value + 1

    def test_on_end_apply_source_transform(self):
        counter_value = COUNTER
        dc_hooks.on_end_apply_source_transform = increase_counter_hook_return_only_second_arg
        dtp = self.create_transformation_pipeline()
        dtp.execute()
        assert COUNTER == counter_value + 1


import os
import shutil
import unittest
from copy import deepcopy
from typing import Optional, Tuple, Any, Callable, Dict, Sequence, List

import pandas as pd
from pandas.testing import assert_frame_equal

from datacode.models.column.column import Column
from datacode.models.dtypes.str_type import StringType
from datacode.models.source import DataSource
from datacode.models.variables import Variable
from datacode.models.variables.expression import Expression
from datacode import Transform, DataOutputNotSafeException, Index, ColumnIndex
from tests.utils import GENERATED_PATH, assert_frame_not_equal


def transform_cell_data_func(col: Column, variable: Variable, cell: Any) -> Any:
    if isinstance(cell, str):
        return cell

    return cell + 1


def transform_series_data_func(col: Column, variable: Variable, series: pd.Series) -> pd.Series:
    return series + 1


def transform_dataframe_data_func(col: Column, variable: Variable, df: pd.DataFrame) -> pd.DataFrame:
    df[variable.name] = df[variable.name] + 1
    return df


def transform_source_data_func(col: Column, variable: Variable, source: DataSource) -> DataSource:
    # Extra unnecessary logic to access source.columns to test looking up columns
    cols = source.columns
    for this_col in cols:
        if not this_col.variable.key == col.variable.key:
            continue
        if not variable.dtype.is_numeric:
            continue
        source.df[variable.name] = source.df[variable.name] + 1
    return source


def expression_series_func(cols: Sequence[Column]) -> pd.Series:
    return cols[0].series + cols[1].series


class SourceTest(unittest.TestCase):
    test_df = pd.DataFrame(
        [
            (1, 2, 'd'),
            (3, 4, 'd'),
            (5, 6, 'e')
        ],
        columns=['a', 'b', 'c']
    )
    expect_df_no_rename_a_plus_1 = pd.DataFrame(
        [
            (2, 2, 'd'),
            (4, 4, 'd'),
            (6, 6, 'e')
        ],
        columns=['a', 'b', 'c']
    )
    expect_loaded_df_rename_only = pd.DataFrame(
        [
            (1, 2, 'd'),
            (3, 4, 'd'),
            (5, 6, 'e')
        ],
        columns=['A', 'B', 'C'],
    ).convert_dtypes()
    expect_loaded_df_rename_only_indexed_c = expect_loaded_df_rename_only.set_index('C')
    expect_loaded_df_rename_only_a_indexed_c = expect_loaded_df_rename_only_indexed_c.drop('B', axis=1)
    expect_loaded_df_rename_only_a_b = pd.DataFrame(
        [
            (1, 2,),
            (3, 4,),
            (5, 6,)
        ],
        columns=['A', 'B']
    ).convert_dtypes()
    expect_loaded_df_with_transform = pd.DataFrame(
        [
            (2, 3, 'd'),
            (4, 5, 'd'),
            (6, 7, 'e')
        ],
        columns=['A_1', 'B_1', 'C']
    ).convert_dtypes()
    expect_loaded_df_with_a_and_a_transform = pd.DataFrame(
        [
            (1, 2, 2, 'd'),
            (3, 4, 4, 'd'),
            (5, 6, 6, 'e')
        ],
        columns=['A', 'A_1', 'B', 'C']
    ).convert_dtypes()
    expect_loaded_df_with_a_transform_and_a = pd.DataFrame(
        [
            (2, 1, 2, 'd'),
            (4, 3, 4, 'd'),
            (6, 5, 6, 'e')
        ],
        columns=['A_1', 'A', 'B', 'C']
    ).convert_dtypes()
    expect_loaded_df_with_transform_only_a_b = pd.DataFrame(
        [
            (2, 3,),
            (4, 5,),
            (6, 7,)
        ],
        columns=['A_1', 'B_1']
    ).convert_dtypes()
    expect_loaded_df_with_transform_and_a_pre_transformed = pd.DataFrame(
        [
            (1, 3, 'd'),
            (3, 5, 'd'),
            (5, 7, 'e')
        ],
        columns=['A_1', 'B_1', 'C']
    ).convert_dtypes()
    expect_loaded_df_with_calculated = pd.DataFrame(
        [
            (1, 2, 'd', 3),
            (3, 4, 'd', 7),
            (5, 6, 'e', 11)
        ],
        columns=['A', 'B', 'C', 'D'],
    ).convert_dtypes()
    expect_loaded_df_with_calculated_c_d_only = pd.DataFrame(
        [
            ('d', 3),
            ('d', 7),
            ('e', 11)
        ],
        columns=['C', 'D'],
    ).convert_dtypes()
    expect_loaded_df_with_calculated_transformed = pd.DataFrame(
        [
            (1, 2, 'd', 4),
            (3, 4, 'd', 8),
            (5, 6, 'e', 12)
        ],
        columns=['A', 'B', 'C', 'D_1'],
    ).convert_dtypes()
    expect_loaded_df_with_calculated_and_calculated_transformed = pd.DataFrame(
        [
            (1, 2, 'd', 3, 4),
            (3, 4, 'd', 7, 8),
            (5, 6, 'e', 11, 12)
        ],
        columns=['A', 'B', 'C', 'D', 'D_1'],
    ).convert_dtypes()
    expect_loaded_df_with_calculated_transformed_and_calculated = pd.DataFrame(
        [
            (1, 2, 'd', 4, 3),
            (3, 4, 'd', 8, 7),
            (5, 6, 'e', 12, 11)
        ],
        columns=['A', 'B', 'C', 'D_1', 'D'],
    ).convert_dtypes()
    expect_loaded_df_with_calculate_on_transformed_before_transform = pd.DataFrame(
        [
            (2, 3, 'd', 3),
            (4, 5, 'd', 7),
            (6, 7, 'e', 11)
        ],
        columns=['A_1', 'B_1', 'C', 'D'],
    ).convert_dtypes()
    expect_loaded_df_with_calculate_on_transformed_after_transform = pd.DataFrame(
        [
            (2, 3, 'd', 5),
            (4, 5, 'd', 9),
            (6, 7, 'e', 13)
        ],
        columns=['A_1', 'B_1', 'C', 'D'],
    ).convert_dtypes()
    expect_loaded_df_with_calculate_on_transformed_before_and_after_transform = pd.DataFrame(
        [
            (2, 3, 'd', 4),
            (4, 5, 'd', 8),
            (6, 7, 'e', 12)
        ],
        columns=['A_1', 'B_1', 'C', 'D'],
    ).convert_dtypes()
    expect_loaded_df_categorical = expect_loaded_df_rename_only.copy()
    expect_loaded_df_categorical['C'] = expect_loaded_df_categorical['C'].astype('category')
    transform_name_func = lambda x: f'{x}_1'
    transform_cell = Transform('add_one_cell', transform_name_func, transform_cell_data_func, data_func_target='cell')
    transform_series = Transform('add_one_series', transform_name_func, transform_series_data_func, data_func_target='series')
    transform_dataframe = Transform('add_one_df', transform_name_func, transform_dataframe_data_func, data_func_target='dataframe')
    transform_source = Transform('add_one_source', transform_name_func, transform_source_data_func, data_func_target='source')
    csv_path = os.path.join(GENERATED_PATH, 'data.csv')

    def setup_method(self, *args, **kwargs):
        if os.path.exists(GENERATED_PATH):
            shutil.rmtree(GENERATED_PATH)
        os.makedirs(GENERATED_PATH)

    def teardown_method(self, *args, **kwargs):
        shutil.rmtree(GENERATED_PATH)

    def create_source(self, **kwargs) -> DataSource:
        config_dict = dict(
            df=self.test_df,
            location=self.csv_path,
        )
        config_dict.update(kwargs)
        return DataSource(**config_dict)

    def get_transform(self, func_type: str) -> Transform:
        if func_type == 'cell':
            return self.transform_cell
        elif func_type == 'series':
            return self.transform_series
        elif func_type == 'dataframe':
            return self.transform_dataframe
        elif func_type == 'source':
            return self.transform_source
        else:
            raise ValueError(
                f'could not look up func_type {func_type}, should be one of cell, series, dataframe, source')

    def create_csv(self, df: Optional[pd.DataFrame] = None, **to_csv_kwargs):
        if df is None:
            df = self.test_df
        df.to_csv(self.csv_path, index=False, **to_csv_kwargs)

    def get_transform_dict(self, transform_data: str = '', apply_transforms: bool = True):
        if transform_data:
            transform = self.get_transform(transform_data)
            transform_dict = dict(
                available_transforms=[transform],
            )
            if apply_transforms:
                transform_dict['applied_transforms'] = [transform]
        else:
            transform_dict = {}

        return transform_dict

    def create_variables(self, transform_data: str = '', apply_transforms: bool = True) -> Tuple[Variable, Variable, Variable]:
        transform_dict = self.get_transform_dict(transform_data=transform_data, apply_transforms=apply_transforms)

        a = Variable('a', 'A', dtype='int', **transform_dict)
        b = Variable('b', 'B', dtype='int', **transform_dict)
        c = Variable('c', 'C', dtype='str')
        return a, b, c

    def create_columns(self, transform_data: str = '', apply_transforms: bool = True) -> List[Column]:
        a, b, c = self.create_variables(transform_data=transform_data, apply_transforms=apply_transforms)
        ac = Column(a, 'a')
        bc = Column(b, 'b')
        cc = Column(c, 'c')
        return [
            ac,
            bc,
            cc
        ]

    def create_c_index(self) -> Index:
        c_index = Index('c', dtype=StringType(categorical=True))
        return c_index

    def create_variables_and_c_colindex(self, transform_data: str = '', apply_transforms: bool = True
                                        ) -> Tuple[List[Variable], ColumnIndex]:
        a, b, c = self.create_variables(transform_data=transform_data, apply_transforms=apply_transforms)
        c_index = self.create_c_index()

        c_col_index = ColumnIndex(c_index, [c])

        return [a, b, c], c_col_index

    def create_indexed_columns(self, transform_data: str = '', apply_transforms: bool = True) -> List[Column]:
        (a, b, c), c_col_index = self.create_variables_and_c_colindex(
            transform_data=transform_data, apply_transforms=apply_transforms
        )
        ac = Column(a, 'a', indices=[c_col_index])
        bc = Column(b, 'b', indices=[c_col_index])
        cc = Column(c, 'c')
        return [
            ac,
            bc,
            cc
        ]


class TestCreateSource(SourceTest):

    def test_create_source_from_df(self):
        ds = self.create_source(location=None)
        assert_frame_equal(ds.df, self.test_df)

    def test_create_source_from_file_path(self):
        self.create_csv()
        ds = self.create_source(df=None)
        assert_frame_equal(ds.df, self.test_df)

    def test_create_source_with_columns(self):
        all_cols = self.create_columns()
        ds = self.create_source(location=None, columns=all_cols)
        assert ds.columns == all_cols

    def test_graph(self):
        ds = self.create_source(location=None)
        # Need to do a better job with this test, see TestDataMergePipeline.test_graph
        ds.graph


class TestLoadSource(SourceTest):

    def test_load_with_columns(self):
        self.create_csv()
        all_cols = self.create_columns()
        ds = self.create_source(df=None, columns=all_cols)
        assert_frame_equal(ds.df, self.expect_loaded_df_rename_only)

    def test_load_with_columns_subset(self):
        self.create_csv()
        all_cols = self.create_columns()
        all_vars = self.create_variables()
        var_subset = [var for var in all_vars if var.key != 'c']
        ds = self.create_source(df=None, columns=all_cols, load_variables=var_subset)
        assert_frame_equal(ds.df, self.expect_loaded_df_rename_only_a_b)

    def test_with_with_repeated_variables_different_transforms(self):
        self.create_csv()
        all_cols = self.create_columns(transform_data='cell', apply_transforms=False)
        a, b, c = self.create_variables(transform_data='cell', apply_transforms=False)

        # First with original variable first, then transformation
        load_variables = [
            a,
            a.add_one_cell(),
            b,
            c,
        ]
        ds = self.create_source(df=None, columns=all_cols, load_variables=load_variables)
        assert_frame_equal(ds.df, self.expect_loaded_df_with_a_and_a_transform)

        all_cols = self.create_columns(transform_data='cell', apply_transforms=False)
        a, b, c = self.create_variables(transform_data='cell', apply_transforms=False)
        # Now with transformation first, then original variable
        load_variables = [
            a.add_one_cell(),
            a,
            b,
            c,
        ]
        ds = self.create_source(df=None, columns=all_cols, load_variables=load_variables)
        assert_frame_equal(ds.df, self.expect_loaded_df_with_a_transform_and_a)

    def test_load_with_repeated_variable_names_raises_error(self):
        self.create_csv()
        all_cols = self.create_columns()
        all_cols.append(deepcopy(all_cols[2]))
        with self.assertRaises(ValueError) as cm:
            ds = self.create_source(df=None, columns=all_cols)
            exc = cm.exception
            assert 'variable name C repeated in load variables' in str(exc)

    def test_with_columns_and_load_variables_with_transforms(self):
        self.create_csv()
        all_cols = self.create_columns(transform_data='cell', apply_transforms=False)
        a, b, c = self.create_variables(transform_data='cell', apply_transforms=False)
        load_variables = [
            a.add_one_cell(),
            b.add_one_cell(),
            c
        ]
        ds = self.create_source(df=None, columns=all_cols, load_variables=load_variables)
        assert_frame_equal(ds.df, self.expect_loaded_df_with_transform)

    def test_load_with_columns_and_transform_cell(self):
        self.create_csv()
        all_cols = self.create_columns(transform_data='cell')
        ds = self.create_source(df=None, columns=all_cols)
        assert_frame_equal(ds.df, self.expect_loaded_df_with_transform)

    def test_load_with_columns_and_transform_series(self):
        self.create_csv()
        all_cols = self.create_columns(transform_data='series')
        ds = self.create_source(df=None, columns=all_cols)
        assert_frame_equal(ds.df, self.expect_loaded_df_with_transform)

    def test_load_with_columns_and_transform_dataframe(self):
        self.create_csv()
        all_cols = self.create_columns(transform_data='dataframe')
        ds = self.create_source(df=None, columns=all_cols)
        assert_frame_equal(ds.df, self.expect_loaded_df_with_transform)

    def test_load_with_columns_and_transform_source(self):
        self.create_csv()
        all_cols = self.create_columns(transform_data='source')
        ds = self.create_source(df=None, columns=all_cols)
        assert_frame_equal(ds.df, self.expect_loaded_df_with_transform)

    def test_load_with_columns_transforms_and_pre_applied_transforms(self):
        self.create_csv()
        all_cols = self.create_columns(transform_data='cell')
        a, b, c = self.create_variables(transform_data='cell')
        all_cols[0] = Column(a, 'a', applied_transform_keys=['add_one_cell'])
        ds = self.create_source(df=None, columns=all_cols)
        assert_frame_equal(ds.df, self.expect_loaded_df_with_transform_and_a_pre_transformed)

    def test_load_with_categorical(self):
        self.create_csv()
        all_cols = self.create_columns()
        a, b, c = self.create_variables()
        all_cols[2] = Column(c, 'c', dtype=StringType(categorical=True))
        ds = self.create_source(df=None, columns=all_cols)
        assert_frame_equal(ds.df, self.expect_loaded_df_categorical)

    def test_load_with_datetime(self):
        test_df = self.test_df.copy()
        test_df['d'] = pd.to_datetime('1/1/2000')
        self.create_csv(df=test_df)

        expect_df = self.expect_loaded_df_rename_only.copy()
        expect_df['Date'] = pd.to_datetime('1/1/2000')

        date_var = Variable('Date', dtype='datetime')
        date_col = Column(date_var, 'd')
        all_cols = self.create_columns()
        all_cols.append(date_col)

        ds = self.create_source(df=None, columns=all_cols)
        assert_frame_equal(ds.df, expect_df)

    def test_load_with_calculated_variable(self):
        self.create_csv()
        all_cols = self.create_columns()
        a, b, c = self.create_variables()
        d = Variable('d', 'D', calculation=a + b)
        ds = self.create_source(df=None, columns=all_cols, load_variables=[a, b, c, d])
        assert_frame_equal(ds.df, self.expect_loaded_df_with_calculated)

    def test_load_with_calculated_variable_from_func(self):
        self.create_csv()
        all_cols = self.create_columns()
        a, b, c = self.create_variables()
        expr = Expression([a, b], func=expression_series_func, summary='Add em up')
        d = Variable('d', 'D', calculation=expr)
        ds = self.create_source(df=None, columns=all_cols, load_variables=[a, b, c, d])
        assert_frame_equal(ds.df, self.expect_loaded_df_with_calculated)

    def test_load_with_calculated_transformed_variable(self):
        self.create_csv()
        all_cols = self.create_columns()
        a, b, c = self.create_variables()
        tran = self.get_transform('cell')
        d = Variable('d', 'D', calculation=a + b, available_transforms=[tran])
        ds = self.create_source(df=None, columns=all_cols, load_variables=[a, b, c, d.add_one_cell()])
        assert_frame_equal(ds.df, self.expect_loaded_df_with_calculated_transformed)

    def test_load_with_calculated_and_same_calculated_variable_transformed(self):
        self.create_csv()

        # Try with plain calculated variable first
        all_cols = self.create_columns()
        a, b, c = self.create_variables()
        tran = self.get_transform('cell')
        d = Variable('d', 'D', calculation=a + b, available_transforms=[tran])
        load_vars = [
            a,
            b,
            c,
            d,
            d.add_one_cell()
        ]
        ds = self.create_source(df=None, columns=all_cols, load_variables=load_vars)
        assert_frame_equal(ds.df, self.expect_loaded_df_with_calculated_and_calculated_transformed)

        # Now try with plain calculated variable second
        all_cols = self.create_columns()
        a, b, c = self.create_variables()
        tran = self.get_transform('cell')
        d = Variable('d', 'D', calculation=a + b, available_transforms=[tran])
        load_vars = [
            a,
            b,
            c,
            d.add_one_cell(),
            d,
        ]
        ds = self.create_source(df=None, columns=all_cols, load_variables=load_vars)
        assert_frame_equal(ds.df, self.expect_loaded_df_with_calculated_transformed_and_calculated)

    def test_load_with_calculate_on_transformed_after_transform(self):
        self.create_csv()
        all_cols = self.create_columns()
        a, b, c = self.create_variables(transform_data='cell')
        d = Variable('d', 'D', calculation=a + b)
        ds = self.create_source(df=None, columns=all_cols, load_variables=[a, b, c, d])
        assert_frame_equal(ds.df, self.expect_loaded_df_with_calculate_on_transformed_after_transform)

    def test_load_with_calculate_on_transformed_before_transform(self):
        self.create_csv()
        all_cols = self.create_columns()
        a, b, c = self.create_variables(transform_data='cell', apply_transforms=False)
        d = Variable('d', 'D', calculation=a + b)
        ds = self.create_source(df=None, columns=all_cols, load_variables=[a.add_one_cell(), b.add_one_cell(), c, d])
        assert_frame_equal(ds.df, self.expect_loaded_df_with_calculate_on_transformed_before_transform)

    def test_load_with_calculate_on_transformed_before_and_after_transform(self):
        self.create_csv()
        all_cols = self.create_columns()
        a, b, c = self.create_variables(transform_data='cell', apply_transforms=False)
        d = Variable('d', 'D', calculation=a + b.add_one_cell())
        ds = self.create_source(df=None, columns=all_cols, load_variables=[a.add_one_cell(), b.add_one_cell(), c, d])
        assert_frame_equal(ds.df, self.expect_loaded_df_with_calculate_on_transformed_before_and_after_transform)

    def test_load_with_calculated_variable_using_non_passed_load_variables(self):
        self.create_csv()
        all_cols = self.create_columns()
        a, b, c = self.create_variables()
        d = Variable('d', 'D', calculation=a + b)
        ds = self.create_source(df=None, columns=all_cols, load_variables=[c, d])
        assert_frame_equal(ds.df, self.expect_loaded_df_with_calculated_c_d_only)


class TestDunders(SourceTest):

    def test_str(self):
        source = self.create_source(location=None)
        print(source)


class TestTransform(SourceTest):

    def test_transform_existing_source(self):
        self.create_csv()
        all_cols = self.create_columns()
        a, b, c = self.create_variables()
        orig_ds = self.create_source(df=None, columns=all_cols, load_variables=[a, b])
        all_ds = []
        all_ds.append(self.transform_cell.apply_to_source(orig_ds))
        all_ds.append(self.transform_series.apply_to_source(orig_ds))
        all_ds.append(self.transform_dataframe.apply_to_source(orig_ds))
        all_ds.append(self.transform_source.apply_to_source(orig_ds))
        for ds in all_ds:
            assert_frame_equal(ds.df, self.expect_loaded_df_with_transform_only_a_b)

    def test_transform_subset_existing_source(self):
        self.create_csv()
        all_cols = self.create_columns()
        a, b, c = self.create_variables()
        orig_ds = self.create_source(df=None, columns=all_cols)
        all_ds = []
        all_ds.append(self.transform_cell.apply_to_source(orig_ds, subset=[a, b]))
        all_ds.append(self.transform_series.apply_to_source(orig_ds, subset=[a, b]))
        all_ds.append(self.transform_dataframe.apply_to_source(orig_ds, subset=[a, b]))
        all_ds.append(self.transform_source.apply_to_source(orig_ds, subset=[a, b]))
        for ds in all_ds:
            assert_frame_equal(ds.df, self.expect_loaded_df_with_transform)


class TestOutput(SourceTest):

    def test_save_then_load(self):
        self.create_csv()
        all_cols = self.create_columns()
        ds = self.create_source(df=None, columns=all_cols)
        ds.output()
        ds = self.create_source(df=None, columns=all_cols)
        assert_frame_equal(ds.df, self.expect_loaded_df_rename_only)

    def test_save_then_load_with_indices(self):
        self.create_csv()
        all_cols = self.create_indexed_columns()
        ds = self.create_source(df=None, columns=all_cols)
        ds.output()
        ds = self.create_source(df=None, columns=all_cols)
        assert_frame_equal(ds.df, self.expect_loaded_df_rename_only_indexed_c)

    def test_save_then_load_variable_subset(self):
        self.create_csv()
        all_cols = self.create_columns()
        all_vars = self.create_variables()
        var_subset = [var for var in all_vars if var.key != 'c']

        # Test when in safe mode, raises error because will delete data
        ds = self.create_source(df=None, columns=all_cols, load_variables=var_subset)
        with self.assertRaises(DataOutputNotSafeException) as cm:
            ds.output()

        # Test when bypassing safe mode, can save and load properly
        ds = self.create_source(
            df=None, columns=all_cols, load_variables=var_subset, data_outputter_kwargs=dict(safe=False)
        )
        ds.output()

        ds = self.create_source(df=None, columns=all_cols, load_variables=var_subset)
        assert_frame_equal(ds.df, self.expect_loaded_df_rename_only_a_b)

    def test_save_then_load_variable_subset_and_indices(self):
        self.create_csv()
        all_cols = self.create_indexed_columns()
        (a, b, c), c_col_index = self.create_variables_and_c_colindex()
        var_subset = [a, c]

        # Test when in safe mode, raises error because will delete data
        ds = self.create_source(df=None, columns=all_cols, load_variables=var_subset)
        with self.assertRaises(DataOutputNotSafeException) as cm:
            ds.output()

        # Test when bypassing safe mode, can save and load properly
        ds = self.create_source(
            df=None, columns=all_cols, load_variables=var_subset, data_outputter_kwargs=dict(safe=False)
        )
        ds.output()

        ds = self.create_source(df=None, columns=all_cols, load_variables=var_subset)
        assert_frame_equal(ds.df, self.expect_loaded_df_rename_only_a_indexed_c)

    def test_save_then_load_with_transformations(self):
        self.create_csv()
        all_cols = self.create_columns(transform_data='cell', apply_transforms=False)
        a, b, c = self.create_variables(transform_data='cell', apply_transforms=False)
        load_variables = [
            a.add_one_cell(),
            b.add_one_cell(),
            c
        ]
        ds = self.create_source(df=None, columns=all_cols, load_variables=load_variables)
        ds.output()

        # This works because the same variables and columns are used to load again
        ds = self.create_source(df=None, columns=all_cols, load_variables=load_variables)
        assert_frame_equal(ds.df, self.expect_loaded_df_with_transform)

        # This won't work right, applying the transformations a second time
        all_cols = self.create_columns(transform_data='cell', apply_transforms=False)
        a, b, c = self.create_variables(transform_data='cell', apply_transforms=False)
        load_variables = [
            a.add_one_cell(),
            b.add_one_cell(),
            c
        ]
        ds = self.create_source(df=None, columns=all_cols, load_variables=load_variables)
        assert_frame_not_equal(ds.df, self.expect_loaded_df_with_transform)



from .interface import MatchComparisonBarData, MatchComparisonBarGraphData, IDComparisonCollection
import math

class MatchComparisonBar:

    def __init__(self, data: MatchComparisonBarData):
        self.data = data

class BarText:

    def __init__(self, label, bar_length, total_length, xy=(0,0), **text_kwargs):
        self.label = label
        self.bar_length = bar_length
        self.total_length = total_length
        self.xy = xy
        self.text_kwargs = text_kwargs
        self.shifted = False

    def draw(self, plt=None, bar_num=0):
        """

        :param plt:
        :type plt:
        :param bar_num: position of bar within triple bar. i.e. a single bar has three bars in it. 0 is the left.
            1 is the middle, and 2 is the right.
        :type bar_num: int
        :return:
        :rtype:
        """
        if plt is None:
            import matplotlib.pyplot as plt

        self._set_position(bar_num)
        self._set_text_kwarg_defaults()
        plt.annotate(self.label, xy=self.xy, xytext=self.text_pos, **self.text_kwargs)

    def _set_position(self, bar_num):
        """
        Determine postion of text depending on length of bar and length of text
        :return:
        :rtype:
        """

        # TODO [#2]: handle when labels are so large that they overlap due to not enough x shift.
        #
        # E.g. `MatchComparisonBarData(100000000, 100000, 100000, name='Unbalanced')`

        # Testing shows it takes about 51 numbers to fill a bar
        max_digits_total_bar = 51
        max_digits_this_bar = math.floor((self.bar_length / self.total_length) * max_digits_total_bar)

        # Label fits -- no need to shift
        if len(self.label) < max_digits_this_bar:
            self.text_pos = self.xy
            # self.shifted remains False
            return

        #Implicit else, label does not fit, need to shift

        shift_multiplier = bar_num - 1 #change to -1, 0, 1 instead of 0, 1, 2
        rl_shift = 0.1 * self.total_length
        x_shift = shift_multiplier * rl_shift

        self.text_pos = (self.xy[0] + x_shift, self.xy[1] - 0.35)
        self.shifted = True

    def _set_text_kwarg_defaults(self):
        # Only add arrow if shifted text
        if self.shifted:
            if 'arrowprops' not in self.text_kwargs:
                self.text_kwargs['arrowprops'] = dict(
                    width=0.01, facecolor='black', shrink=0.05, headwidth=0
                )
        if 'horizontalalignment' not in self.text_kwargs:
            self.text_kwargs['horizontalalignment'] = 'center'
        if 'verticalalignment' not in self.text_kwargs:
            self.text_kwargs['verticalalignment'] = 'center'


class MatchComparisonBarGraph:

    def __init__(self, *bars: [MatchComparisonBarData], plt=None, name=None, relative=True, thickness=0.5):
        if plt is None:
            import matplotlib.pyplot as plt
        self._absolute_bars = bars
        self._relative_bars = _convert_bar_data_to_relative(bars)
        self.plt = plt
        self.thickness = thickness
        self.relative = relative
        self._set_graph_data(name)

    @property
    def relative(self):
        return self._relative

    @relative.setter
    def relative(self, relative):
        if relative:
            self.bars = self._relative_bars
        else:
            self.bars = self._absolute_bars

        self._relative = relative

    def _set_graph_data(self, name):
        self.graph_data = _convert_bar_data_to_bar_graph_data(self.bars, name=name)
        self._set_max_bar_length(self.bars)

    def _set_max_bar_length(self, bars):
        if not bars:
            return 0
        self.max_len = max([sum(bar) for bar in bars])

    def draw(self, plt):

        if not self.graph_data:
            # Don't plot
            return plt

        plt.clf() #clear out any previous figures

        #### TEMP
        # import pdb
        # pdb.set_trace()
        #### END TEMP

        if self.graph_data.name is not None:
            _draw_title(plt, self.graph_data.name)
        _draw_bars(plt, *self.graph_data, thickness=self.thickness)
        _draw_bar_labels(plt, [bar.name for bar in self.bars], self.max_len)
        _draw_number_labels(plt, self._absolute_bars, self.max_len, self.relative)
        plt.axis('off')
        return plt

    @classmethod
    def from_id_comparison_collection(cls, id_comparison: IDComparisonCollection, plt=None):
        if plt is None:
            import matplotlib.pyplot as plt

        if not id_comparison:
            return cls(name=id_comparison.name, plt=plt)
        return cls(*[MatchComparisonBarData.from_id_comparison(compare) for compare in id_comparison], name=id_comparison.name, plt=plt)

def _convert_bar_data_to_bar_graph_data(data: [MatchComparisonBarData], name=None):
    if not data:
        return []
    split_items = [items for items in zip(*data)]
    return MatchComparisonBarGraphData(*split_items, name=name)

def _convert_bar_data_to_relative(data: [MatchComparisonBarData], scale=100):
    out_list = []
    for bardata in data:
        total_items = sum(bardata)
        new_bardata = MatchComparisonBarData(*[(item / total_items) * scale for item in bardata], name=bardata.name)
        out_list.append(new_bardata)

    return out_list

def _draw_bars(plt, left_items, match_items, right_items, thickness=0.5):
    index = [i for i in range(len(left_items))]

    # first bar starts at zero, second bar starts where first left off.
    # then add the left and middle bars to get start for last bar
    new_lefts = [sum(items) for items in zip(*[left_items, match_items])]

    _draw_bars_part(plt, index, left_items, thickness=thickness, left=None)
    _draw_bars_part(plt, index, match_items, thickness=thickness, left=left_items)
    _draw_bars_part(plt, index, right_items, thickness=thickness, left=new_lefts)


def _draw_bars_part(plt, index, items, thickness=0.5, left=None):
    return plt.barh(index, items, thickness, left=left)

def _draw_title(plt, title):
    plt.title(title)

def _draw_bar_labels(plt, labels, bar_max_width=5):
    [plt.text(-0.05 * bar_max_width, height, label, horizontalalignment='right') for height, label in enumerate(labels)]

def _draw_number_labels(plt, bars: [MatchComparisonBarData], max_len, relative=True):
    for i, bar in enumerate(bars):
        for j, num_label in enumerate(bar):
            label = str(num_label)
            sum_bar = sum(bar)
            if relative:
                pos = (bar.midpoints[j]/sum_bar) * max_len
                bar_len = (bar[j]/sum_bar) * max_len
            else:
                pos = bar.midpoints[j]
                bar_len = bar[j]
            text = BarText(label, bar_len, max_len, xy=(pos, i), horizontalalignment='center', verticalalignment='center')
            text.draw(plt, bar_num=j)


import pandas as pd

from datacode.panel.expandselect import expand_entity_date_selections
from datacode.summarize.subset.outliers.typing import (
    StrList,
    AssociatedColDict,
    BoolDict,
    DfDict,
    TwoDfDictAndDfTuple,
    MinMaxDict
)


def outlier_summary_dicts(df: pd.DataFrame, associated_col_dict: AssociatedColDict,
                          min_max_dict: MinMaxDict,
                          ascending_sort_dict: BoolDict = None,
                          always_associated_cols: StrList = None, bad_column_name: str = '_bad_column',
                          num_firms: int = 3, firm_id_col: str = 'TICKER',
                          date_col: str = 'Date',
                          begin_datevar: str = 'Begin Date',
                          end_datevar: str = 'End Date',
                          expand_months: int = 3
                          ) -> TwoDfDictAndDfTuple:

    bad_df = _bad_df_from_df(
        df,
        min_max_dict,
        bad_column_name=bad_column_name
    )

    bad_df_dict = _column_bad_df_dict(
        bad_df,
        associated_col_dict,
        ascending_sort_dict=ascending_sort_dict,
        always_associated_cols=always_associated_cols,
        bad_column_name=bad_column_name
    )

    selected_orig_df_dict = _col_bad_df_dict_to_selected_orig_df_dict(
        df,
        bad_df_dict,
        associated_col_dict,
        always_associated_cols=always_associated_cols,
        num_firms=num_firms,
        firm_id_col=firm_id_col,
        date_col=date_col,
        begin_datevar=begin_datevar,
        end_datevar=end_datevar,
        expand_months=expand_months
    )

    return bad_df_dict, selected_orig_df_dict, bad_df


def drop_outliers_by_cutoffs(df: pd.DataFrame, min_max_dict: MinMaxDict):
    valid_df = df.copy()
    for col, min_max in min_max_dict.items():
        col_min = min_max[0]
        col_max = min_max[1]
        if col in df.columns:
            valid_df = valid_df.loc[
                (valid_df[col] >= col_min) & (valid_df[col] <= col_max)
            ]


    return valid_df


def _bad_df_from_df(df: pd.DataFrame, min_max_dict: MinMaxDict,
             bad_column_name: str = '_bad_column'):
    bad_df = pd.DataFrame()
    for col, min_max in min_max_dict.items():
        col_min = min_max[0]
        col_max = min_max[1]
        col_bad_df = df.loc[
            (df[col] < col_min) | (df[col] > col_max)
            ]
        col_bad_df[bad_column_name] = col
        bad_df = bad_df.append(
            col_bad_df
        )

    return bad_df

def _column_bad_df_dict(bad_df: pd.DataFrame, associated_col_dict: AssociatedColDict,
                        ascending_sort_dict: BoolDict = None,
                        always_associated_cols: StrList = None, bad_column_name: str = '_bad_column') -> DfDict:
    if always_associated_cols is None:
        always_associated_cols = []

    # Default is ascending=False for all variables
    if ascending_sort_dict is None:
        ascending_sort_dict = {col: False for col in associated_col_dict}

    df_dict = {}
    for col, associated_cols in associated_col_dict.items():
        df_dict[col] = _sorted_bad_df_for_column(
            bad_df,
            col,
            always_associated_cols + associated_cols,
            bad_column_name=bad_column_name,
            sort_ascending=ascending_sort_dict[col]
        )

    return df_dict


def _col_bad_df_dict_to_selected_orig_df_dict(df: pd.DataFrame, col_bad_df_dict: DfDict,
                                              associated_col_dict: AssociatedColDict,
                                              always_associated_cols: StrList = None,
                                              num_firms: int = 3, firm_id_col: str = 'TICKER',
                                              date_col: str = 'Date',
                                              begin_datevar: str = 'Begin Date',
                                              end_datevar: str = 'End Date',
                                              expand_months: int = 3) -> DfDict:
    df_dict = {}
    for col, col_bad_df in col_bad_df_dict.items():
        df_dict[col] = expand_entity_date_selections(
            df,
            col_bad_df,
            [col] + always_associated_cols + associated_col_dict[col],
            num_firms=num_firms,
            expand_months=expand_months,
            entity_id_col=firm_id_col,
            date_col=date_col,
            begin_datevar=begin_datevar,
            end_datevar=end_datevar
        )

    return df_dict


def _sorted_bad_df_for_column(bad_df: pd.DataFrame, col: str, associated_cols: StrList,
                              bad_column_name: str = '_bad_column', sort_ascending: bool = False) -> pd.DataFrame:
    return bad_df.loc[
        bad_df[bad_column_name] == col,
        associated_cols + [col]
    ].sort_values(col, ascending=sort_ascending)






class NamedBase:

    def __repr__(self):
        cls_name = self.__class__.__name__
        return f'<{cls_name}(name={self.name}, items={[item for item in self]})>'

class NamedSet(NamedBase, set):

    def __init__(self, iterable=None, name=None):
        self.name = name
        super().__init__(iterable)


class NamedList(NamedBase, list):

    def __init__(self, iterable=None, name=None):
        self.name = name
        super().__init__(iterable)


from typing import Tuple

from pandas.testing import assert_frame_equal
import pandas as pd
import numpy as np

from datacode import VariableCollection, Variable, Column, Index, StringType, ColumnIndex, SourceTransform
from tests.test_source import SourceTest
from tests.test_variables import VC_NAME


class SpecificTransformsTest(SourceTest):
    time_index = Index('time', dtype='datetime')
    by_index = Index('id', dtype=StringType(categorical=True))
    test_df_with_ids_and_dates = pd.DataFrame(
        [
            (1, 2, 'd'),
            (2, 4, 'd'),
            (3, 6, 'd'),
            (4, 8, 'e'),
            (5, 10, 'e'),
            (6, 12, 'e'),
        ],
        columns=['a', 'b', 'c']
    )
    orig_date_index = pd.date_range(start='1/1/2000', periods=3, freq='d')
    date_index_with_gaps = pd.Index.union(pd.DatetimeIndex([pd.to_datetime('1/1/2000')]),
                                     pd.date_range(start='1/3/2000', periods=2, freq='d'))
    full_date_index = pd.Index.append(orig_date_index, date_index_with_gaps)
    test_df_with_ids_and_dates['date'] = full_date_index
    c_df_index = pd.Index(['d', 'd', 'e'], name='C')
    expect_loaded_df_with_lags = pd.DataFrame(
        [
            (np.nan, np.nan, 'd'),
            (1, 2, 'd'),
            (3, 4, 'e')
        ],
        columns=['A$_{t - 1}$', 'B$_{t - 1}$', 'C'],
    ).convert_dtypes()
    expect_loaded_df_with_two_lags = pd.DataFrame(
        [
            (np.nan, np.nan, 'd'),
            (np.nan, np.nan, 'd'),
            (1, 2, 'e')
        ],
        columns=['A$_{t - 2}$', 'B$_{t - 2}$', 'C'],
    ).convert_dtypes()
    expect_loaded_df_with_lags_and_by_var = pd.DataFrame(
        [
            (np.nan, np.nan),
            (1, 2),
            (np.nan, np.nan)
        ],
        columns=['A$_{t - 1}$', 'B$_{t - 1}$'],
        index=c_df_index,
    ).convert_dtypes()
    expect_lag_df_with_ids_and_dates = pd.DataFrame(
        [
            (np.nan, np.nan, 'd'),
            (1, 2, 'd'),
            (2, 4, 'd'),
            (np.nan, np.nan, 'e'),
            (np.nan, np.nan, 'e'),
            (5, 10, 'e'),
        ],
        columns=['A$_{t - 1}$', 'B$_{t - 1}$', 'C'],
    ).convert_dtypes()
    expect_lag_df_with_ids_and_dates['Date'] = full_date_index
    expect_lag_df_with_ids_and_dates.set_index(['C', 'Date'], inplace=True)
    expect_loaded_df_with_change = pd.DataFrame(
        [
            (np.nan, np.nan, 'd'),
            (2, 2, 'd'),
            (2, 2, 'e')
        ],
        columns=['A Change', 'B Change', 'C'],
    ).convert_dtypes()
    expect_loaded_df_with_dual_change = pd.DataFrame(
        [
            (np.nan, np.nan, 'd'),
            (np.nan, np.nan, 'd'),
            (0, 0, 'e')
        ],
        columns=['A Change Change', 'B Change Change', 'C'],
    ).convert_dtypes()
    expect_change_df_with_ids_and_dates = pd.DataFrame(
        [
            (np.nan, np.nan, 'd'),
            (1, 2, 'd'),
            (1, 2, 'd'),
            (np.nan, np.nan, 'e'),
            (np.nan, np.nan, 'e'),
            (1, 2, 'e'),
        ],
        columns=['A Change', 'B Change', 'C'],
    ).convert_dtypes()
    expect_change_df_with_ids_and_dates['Date'] = full_date_index
    expect_change_df_with_ids_and_dates.set_index(['C', 'Date'], inplace=True)


    def create_variable_collection(self, **kwargs) -> Tuple[VariableCollection, Variable, Variable, Variable]:
        config_dict = dict(
            name=VC_NAME
        )
        config_dict.update(**kwargs)
        a, b, c = self.create_variables()
        vc = VariableCollection(a, b, c, **config_dict)
        return vc, a, b, c


class TestLag(SpecificTransformsTest):

    def test_lag_with_defaults_no_indices(self):
        vc, a, b, c = self.create_variable_collection()
        self.create_csv()
        all_cols = self.create_columns()
        load_variables = [
            vc.a.lag(),
            vc.b.lag(),
            c
        ]
        ds = self.create_source(df=None, columns=all_cols, load_variables=load_variables)
        assert_frame_equal(ds.df, self.expect_loaded_df_with_lags)
        assert str(vc.a.lag().symbol) == r'\text{A}_{t - 1}'
        assert str(vc.b.lag().symbol) == r'\text{B}_{t - 1}'
        assert str(vc.c.symbol) == r'\text{C}'

    def test_two_lags_no_indices(self):
        vc, a, b, c = self.create_variable_collection()
        self.create_csv()
        all_cols = self.create_columns()
        load_variables = [
            vc.a.lag(2),
            vc.b.lag(2),
            c
        ]
        ds = self.create_source(df=None, columns=all_cols, load_variables=load_variables)
        assert_frame_equal(ds.df, self.expect_loaded_df_with_two_lags)
        assert str(vc.a.lag(2).symbol) == r'\text{A}_{t - 2}'
        assert str(vc.b.lag(2).symbol) == r'\text{B}_{t - 2}'
        assert str(vc.c.symbol) == r'\text{C}'

    def test_two_separate_lags_no_indices(self):
        vc, a, b, c = self.create_variable_collection()
        self.create_csv()
        all_cols = self.create_columns()
        load_variables = [
            vc.a.lag().lag(),
            vc.b.lag().lag(),
            c
        ]
        ds = self.create_source(df=None, columns=all_cols, load_variables=load_variables)
        assert_frame_equal(ds.df, self.expect_loaded_df_with_two_lags)
        assert str(vc.a.lag().lag().symbol) == r'\text{A}_{t - 2}'
        assert str(vc.b.lag().lag().symbol) == r'\text{B}_{t - 2}'
        assert str(vc.c.symbol) == r'\text{C}'

    def test_lags_with_by_index(self):
        vc, a, b, c = self.create_variable_collection()
        self.create_csv()
        by_colindex = [ColumnIndex(self.by_index, [c])]
        ac = Column(a, 'a', by_colindex)
        bc = Column(b, 'b', by_colindex)
        cc = Column(c, 'c')
        all_cols = [
            ac, bc, cc
        ]
        load_variables = [
            vc.a.lag(),
            vc.b.lag(),
            c
        ]
        ds = self.create_source(df=None, columns=all_cols, load_variables=load_variables)
        assert_frame_equal(ds.df, self.expect_loaded_df_with_lags_and_by_var)

    def test_lags_with_by_index_and_time_index_with_gaps(self):
        vc, a, b, c = self.create_variable_collection()
        d = Variable('d', 'Date', dtype='datetime')
        self.create_csv(df=self.test_df_with_ids_and_dates)
        by_colindex = ColumnIndex(self.by_index, [c])
        time_colindex = ColumnIndex(self.time_index, [d])
        by_time_colindex = [by_colindex, time_colindex]
        ac = Column(a, 'a', by_time_colindex)
        bc = Column(b, 'b', by_time_colindex)
        cc = Column(c, 'c')
        dd = Column(d, 'date')
        all_cols = [
            ac, bc, cc, dd
        ]
        load_variables = [
            vc.a.lag(fill_method=None),
            vc.b.lag(fill_method=None),
            c,
            d
        ]
        ds = self.create_source(df=None, columns=all_cols, load_variables=load_variables)
        assert_frame_equal(ds.df, self.expect_lag_df_with_ids_and_dates)

    def test_lags_as_source_transform_with_subset(self):
        vc, a, b, c = self.create_variable_collection()
        self.create_csv()
        all_cols = self.create_columns()
        load_variables = [
            vc.a,
            vc.b,
            c
        ]
        ds = self.create_source(df=None, columns=all_cols, load_variables=load_variables)
        lag_transform = [transform for transform in vc.transforms if transform.key == 'lag'][0]
        source_transform = SourceTransform.from_transform(lag_transform, subset=[a, b])

        # NOTE: this commented code catches issue where original variables are not preserved when passing
        # preserve_original=False. See SourceTransform.apply

        # source_transform.apply(ds, preserve_original=False)

        ds = source_transform.apply(ds)
        assert_frame_equal(ds.df, self.expect_loaded_df_with_lags)
        assert str(vc.a.lag().symbol) == r'\text{A}_{t - 1}'
        assert str(vc.b.lag().symbol) == r'\text{B}_{t - 1}'
        assert str(vc.c.symbol) == r'\text{C}'


class TestChange(SpecificTransformsTest):

    def test_change_with_defaults_no_indices(self):
        vc, a, b, c = self.create_variable_collection()
        self.create_csv()
        all_cols = self.create_columns()
        load_variables = [
            vc.a.change(),
            vc.b.change(),
            c
        ]
        ds = self.create_source(df=None, columns=all_cols, load_variables=load_variables)
        assert_frame_equal(ds.df, self.expect_loaded_df_with_change)
        assert str(vc.a.change().symbol) == r'\delta \text{A}'
        assert str(vc.b.change().symbol) == r'\delta \text{B}'
        assert str(vc.c.symbol) == r'\text{C}'

    def test_two_separate_changes_no_indices(self):
        vc, a, b, c = self.create_variable_collection()
        self.create_csv()
        all_cols = self.create_columns()
        load_variables = [
            vc.a.change().change(),
            vc.b.change().change(),
            c
        ]
        ds = self.create_source(df=None, columns=all_cols, load_variables=load_variables)
        assert_frame_equal(ds.df, self.expect_loaded_df_with_dual_change)
        assert str(vc.a.change().change().symbol) == r'\delta \delta \text{A}'
        assert str(vc.b.change().change().symbol) == r'\delta \delta \text{B}'
        assert str(vc.c.symbol) == r'\text{C}'

    def test_change_with_by_index_and_time_index_with_gaps(self):
        vc, a, b, c = self.create_variable_collection()
        d = Variable('d', 'Date', dtype='datetime')
        self.create_csv(df=self.test_df_with_ids_and_dates)
        by_colindex = ColumnIndex(self.by_index, [c])
        time_colindex = ColumnIndex(self.time_index, [d])
        by_time_colindex = [by_colindex, time_colindex]
        ac = Column(a, 'a', by_time_colindex)
        bc = Column(b, 'b', by_time_colindex)
        cc = Column(c, 'c')
        dd = Column(d, 'date')
        all_cols = [
            ac, bc, cc, dd
        ]
        load_variables = [
            vc.a.change(fill_method=None),
            vc.b.change(fill_method=None),
            c,
            d
        ]
        ds = self.create_source(df=None, columns=all_cols, load_variables=load_variables)
        assert_frame_equal(ds.df, self.expect_change_df_with_ids_and_dates)


from copy import deepcopy
from typing import Sequence, Optional, List

from mixins import ReprMixin

from datacode.models.variables.tools import _get_obj_or_attr
from datacode.models.transform.transform import Transform
from datacode.models.transform.applied import AppliedTransform
from datacode.models.transform.specific import DEFAULT_TRANSFORMS
from .variable import Variable


class VariableCollection(ReprMixin):
    repr_cols = ['name', 'keys']

    _default_attr = 'obj'

    def __init__(self, *variables: Variable, name: str = 'variables', default_attr: str = 'obj',
                 transforms: Optional[Sequence[Transform]] = DEFAULT_TRANSFORMS,
                 default_transforms: Optional[Sequence[AppliedTransform]] = None):
        if transforms is None:
            transforms = []
        if default_transforms is None:
            default_transforms = []
        self.transforms = transforms
        self.default_transforms = default_transforms
        self.items: [Variable, VariableCollection] = deepcopy(list(variables))
        self.default_attr = default_attr
        self.name = name

    def __contains__(self, item):
        return item in self.items

    @property
    def default_attr(self):
        return self._default_attr

    @default_attr.setter
    def default_attr(self, default_attr):
        self._default_attr = default_attr
        # TODO [#13]: could make variable collection initialization more efficient
        #
        # Currently calling self._set_variables_and_collections() before self._create_variable_map()
        # as variables need to have the custom name attributes created. But then still calling after to
        # set the variables attributes correctly. Could reorganize this.
        self._set_variables_and_collections()
        self.variable_map = self._create_variable_map()
        self._set_variables_and_collections()
        self._set_default_attr_in_nested_collections()

    @property
    def items(self):
        return self._items

    @items.setter
    def items(self, variables):
        # Goes inside any collections which are inside this one to populate all nested variables as well as collections
        self._items = _create_variable_and_collection_list(variables)

    @property
    def keys(self) -> List[str]:
        return [var.key for var in self.variables]

    # Must override the standard list methods which add and remove elements because
    # we must reflect those operations on the variable hash map
    def append(self, item):
        if item not in self.items:
            self.items.append(item)
        self._add_item_to_map(item)

    def extend(self, iterable):
        self.items.extend(iterable)
        self._add_items_to_map(iterable)

    def remove(self, item: Variable):
        self.items.remove(item)
        self._delete_from_map(item.key)

    def pop(self, i):
        variable = self.items.pop(i)
        self._delete_from_map(variable.key)
        return

    def _add_transforms_to_variables(self):
        for item in self.variables:
            for transform in self.transforms:
                item._add_transform(transform)

            item.applied_transforms = []  # reset so that are not applied multiple times
            for applied_transform in self.default_transforms:
                item.applied_transforms.append(applied_transform)
                item._set_name_and_symbol_by_transforms()

    def _add_item_to_map(self, item):
        self.variable_map.update({
            item.name: _get_obj_or_attr(item, self.default_attr)
        })
        self._set_variables_and_collections()

    def _add_items_to_map(self, items):
        self.variable_map.update(_create_variable_map(items, self.default_attr))
        self._set_variables_and_collections()


    def _delete_from_map(self, item_name):
        self._set_variables_and_collections()
        return self.variable_map.pop(item_name)

    def _create_variable_map(self):
        return _create_variable_map(self.items, output=self.default_attr)

    def _get_from_variable_map(self, item):
        try:
            return self.variable_map[item]
        except KeyError:
            raise AttributeError(f'attribute {item} does not exist on {self}')

    def _set_variables_and_collections(self):
        self.variables = [item for item in self.items if isinstance(item, Variable)]
        self._add_transforms_to_variables()
        self.collections = [item for item in self.items if isinstance(item, VariableCollection)]

    def _set_default_attr_in_nested_collections(self):
        """
        When default attr is changed, must reflect that change in lower collections
        """
        for collection in self.collections:
            # Changing default attr one level deep will trigger changes in the next level, and so on (recursive)
            collection.default_attr = self.default_attr

    def __getitem__(self, item):
        if isinstance(item, str):
            return self._get_from_variable_map(item)

        return self.items[item]

    def __getattr__(self, item: str):

        # Implement standard public list methods other than those explicitly defined
        if (not item.startswith('_')) and hasattr(list, item):
            return getattr(self.items, item)

        return self._get_from_variable_map(item)

    def __dir__(self):
        return self.variable_map.keys()

    @classmethod
    def from_display_names(cls, *names):
        variables = [Variable.from_display_name(disp_name) for disp_name in names]
        return cls(*variables, default_attr=cls._default_attr)

    @classmethod
    def from_variable_definition_dict(cls, def_dict, name=None):
        return _definition_dict_to_collection(def_dict, name=name, default_attr=cls._default_attr)



def _create_variable_map(variables, output='obj'):
    tups = _create_variable_tuples(variables, output=output)
    return dict(tups)

def _create_variable_tuples(variables, output='obj', tups=None):
    if tups is None:
        tups = []
    for variable_or_collection in variables:
        if isinstance(variable_or_collection, VariableCollection):
            # Add all variables in collection
            tups += _create_variable_tuples(variable_or_collection.items, output=output, tups=tups)
            # Now add the collection itself
            tups.append((variable_or_collection.name, variable_or_collection))
        elif isinstance(variable_or_collection, Variable):
            # Add single variable
            tups.append((variable_or_collection.key, _get_obj_or_attr(variable_or_collection, output)))
        else:
            raise ValueError(f'must pass iterable of Variable or VariableCollection. {variable_or_collection} has type {type(variable_or_collection)}')

    return tups

def _create_variable_and_collection_list(variables, collected=None):
    if collected is None:
        collected = []

    for variable_or_collection in variables:
        if isinstance(variable_or_collection, VariableCollection):
            # Add all variables in collection
            collected = _create_variable_and_collection_list(variable_or_collection.items, collected=collected)
            # Now add the collection itself
            collected.append(variable_or_collection)
        elif isinstance(variable_or_collection, Variable):
            # Add single variable
            if variable_or_collection not in collected:
                collected.append(variable_or_collection)
        else:
            raise ValueError(f'must pass iterable of Variable or VariableCollection. {variable_or_collection} has type {type(variable_or_collection)}')

    return collected

def _definition_dict_to_collection(def_dict, name=None, collection=None, default_attr='obj'):
    if collection is None:
        collection = []

    for key, value in def_dict.items():
        if isinstance(value, dict):
            # recurse through dicts, each time creating another layer of collection
            collection.append(_definition_dict_to_collection(value, name=key))
        elif isinstance(value, list):
            # add each variable in list to current collection
            add_items = [_definition_dict_tuple_or_str_to_variable(val) for val in value]
            collection.extend(add_items)
        elif isinstance(value, (str, tuple)):
            # if just str or tuple is passed, parse and add directly to collection
            collection.append(_definition_dict_tuple_or_str_to_variable(value))
        else:
            raise ValueError(f'invalid definition dict. only dicts, lists, strs, and tuples are allowed, got type {type(value)} for value {value}')

    return VariableCollection(*collection, name=name, default_attr=default_attr)

def _definition_dict_tuple_or_str_to_variable(value):
    if isinstance(value, str):
        # if just str is passed, str is for display name
        return Variable.from_display_name(value)
    elif isinstance(value, tuple):
        # if tuple is passed, it is tuple of name, display name
        return Variable(key=value[0], name=value[1])
    elif isinstance(value, Variable):
        # variable itself was in data, just return it
        return value
    else:
        raise ValueError(f'must pass str or tuple, got type {type(value)} for value {value}')

class VariableDefaultCollection(VariableCollection):

    def __init__(self, *variables, name='variables'):
        super().__init__(*variables, name=name, default_attr=self._default_attr)

class VariableNameCollection(VariableDefaultCollection):

    _default_attr = 'name'


class VariableDisplayNameCollection(VariableDefaultCollection):

    _default_attr = 'display_name'

class VariableChangeNameCollection(VariableDefaultCollection):

    _default_attr = 'change_name'


class VariablePortfolioNameCollection(VariableDefaultCollection):

    _default_attr = 'port_name'

class VariableChangePortfolioNameCollection(VariableDefaultCollection):

    _default_attr = 'change_port_name'


import os

from pandas.testing import assert_frame_equal

GENERATED_PATH = os.path.join('tests', 'generated_files')

def assert_frame_not_equal(*args, **kwargs):
    try:
        assert_frame_equal(*args, **kwargs)
    except AssertionError:
        # frames are not equal
        pass
    else:
        # frames are equal
        raise AssertionError


from copy import deepcopy
from typing import Optional, Sequence, TYPE_CHECKING, Union

from mixins.repr import ReprMixin
from sympy import Symbol

if TYPE_CHECKING:
    from datacode.models.variables.variable import Variable
    from datacode.models.column.column import Column
    from datacode.models.source import DataSource
    from datacode.models.dtypes.base import DataType

from datacode.models.logic.partial import partial
from datacode.models.variables.compare import functions_are_equal
from datacode.models.variables.typing import StrFunc, ValueFunc, SymbolFunc
import datacode.hooks as hooks


class Transform(ReprMixin):
    """
    Tracks and applies changes to variables for data, name, and symbol together
    """
    repr_cols = ['key', 'name_func', 'data_func', 'symbol_func', 'data_func_target', 'new_dtype']

    def __init__(self, key: str, name_func: StrFunc = None, data_func: ValueFunc = None,
                 symbol_func: SymbolFunc = None,
                 data_func_target: str = 'series', new_dtype: Optional[Union['DataType', str]] = None):
        from datacode.models.dtypes.convert import convert_str_to_data_type_if_necessary
        if new_dtype is not None:
            new_dtype = convert_str_to_data_type_if_necessary(new_dtype)

        # If name func passed but not symbol func, convert name func to symbol func
        if name_func is not None and symbol_func is None:
            def sym_func_from_name_func(sym: Symbol, *args, **kwargs) -> Symbol:
                symbol_str = str(sym)
                new_symbol_str = name_func(symbol_str, *args, **kwargs)
                new_symbol = Symbol(new_symbol_str)
                return new_symbol
            symbol_func = sym_func_from_name_func


        self.key = key
        self.name_func = name_func
        self.data_func = data_func
        self.symbol_func = symbol_func
        self.data_func_target = self._validate_data_func_target(data_func_target)
        self.new_dtype = new_dtype

    def __hash__(self):
        hash_attrs = ['key', 'data_func_target']
        hash_items = tuple([getattr(self, attr) for attr in hash_attrs])
        return hash(hash_items)

    def _validate_data_func_target(self, data_func_target: str):
        data_func_target = data_func_target.lower()
        cell_values = ('cell', 'value', 'item', 'c', 'v')
        series_values = ('series', 's')
        df_values = ('df', 'dataframe')
        source_values = ('ds', 'datasource', 'source')
        if data_func_target in cell_values:
            return 'cell'
        if data_func_target in series_values:
            return 'series'
        if data_func_target in df_values:
            return 'dataframe'
        if data_func_target in source_values:
            return 'source'
        raise ValueError(f'Did not pass appropriate data_func_target to Transform {self}, got {data_func_target} '
                         f'which should be one of cell, series, dataframe, or source')

    def apply_to_source(self, source: 'DataSource', preserve_original: bool = True,
                        subset: Optional[Sequence['Variable']] = None) -> 'DataSource':
        """
        Applies transformation to entire data source

        :param source:
        :param preserve_original: True to copy the source before applying transformations. False will decrease
        memory usage but will cause the original source to be partially modified
        :param subset: Variables to apply the transformation to. Defaults to all variables.
        :return:
        """
        if preserve_original:
            source = deepcopy(source)
        else:
            # Even when not preserving original, don't want to modify original variables or columns
            # as they may be used in other sources

            # NOTE: not working yet, see SourceTransform.apply
            source.load_variables = deepcopy(source.load_variables)
            source.columns = deepcopy(source.columns)

        if subset:
            variables = deepcopy(subset)
        else:
            variables = source.load_variables

        rename_dict = {}
        for selected_var in variables:
            col = source.col_for(var_key=selected_var.key)
            var = col.variable  # Don't use passed variable directly besides key as may be different instance

            # Apply data transformation
            source = self._apply_transform_for_column_and_variable_to_source(source, col, var)

            # Update variable
            orig_name = var.name
            var._add_applied_transform(self)
            new_name = var.name

            # Update column name
            if orig_name != new_name:
                rename_dict[orig_name] = new_name

        if rename_dict:
            source.df.rename(columns=rename_dict, inplace=True)

        return source

    def _apply_transform_for_column_and_variable_to_source(self, source: 'DataSource', column: 'Column',
                                                           variable: 'Variable') -> 'DataSource':
        source, column, variable = hooks.on_begin_apply_variable_transform(self, source, column, variable)
        source = self._apply_data_transform_to_column_and_variable_in_source(source, column, variable)
        source = hooks.on_end_apply_variable_transform(self, source, column, variable)
        return source

    def _apply_data_transform_to_column_and_variable_in_source(self, source: 'DataSource', column: 'Column',
                                                               variable: 'Variable') -> 'DataSource':
        if self.data_func is None:
            return source
        data_func_with_col = partial(self.data_func, column, variable)
        if self.data_func_target == 'cell':
            source.df[variable.name] = source.df[variable.name].apply(data_func_with_col)
        elif self.data_func_target == 'series':
            source.df[variable.name] = data_func_with_col(source.df[variable.name])
        elif self.data_func_target == 'dataframe':
            source.df = data_func_with_col(source.df)
        elif self.data_func_target == 'source':
            source = data_func_with_col(source)
        else:
            raise ValueError(f'Did not pass appropriate data_func_target to Transform {self}, got '
                             f'{self.data_func_target} which should be one of cell, series, dataframe, or source')

        if self.new_dtype is not None:
            dtype = self.new_dtype
        else:
            dtype = column.dtype
        dtype_str = dtype.read_file_arg
        source.df[variable.name] = source.df[variable.name].astype(dtype_str)

        column.series = source.df[variable.name]
        return source

    def __eq__(self, other):
        same = True
        try:
            same = same and self.key == other.key
            same = same and functions_are_equal(self.name_func, other.name_func)
            same = same and functions_are_equal(self.data_func, other.data_func)
            same = same and self.data_func_target == other.data_func_target
            return same
        except AttributeError:
            return False



from datacode.typing import IntOrNone, FloatOrNone, Union

def missing_more_than_str(missing_tolerance: IntOrNone, missing_quantile: FloatOrNone,
                          missing_display_str: str, id_var: str) -> str:
    return f'More than {_t_or_q(missing_tolerance, missing_quantile)} {missing_display_str} {id_var}'

def missing_more_than_pct_str(missing_tolerance: IntOrNone, missing_quantile: FloatOrNone,
                              missing_display_str: str, id_var: str) -> str:
    return f'More than {_t_or_q(missing_tolerance, missing_quantile)} {missing_display_str} {id_var} Percentage'

def missing_tolerance_count_str(missing_tolerance: IntOrNone, missing_quantile: FloatOrNone,
                                missing_display_str: str, id_var: str) -> str:
    return f'{_t_or_q(missing_tolerance, missing_quantile)} or less {missing_display_str} {id_var} Count'

def id_count_str(id_var: str) -> str:
    return f'{id_var} Count'

def num_or_pct(missing_tolerance: IntOrNone, missing_quantile: FloatOrNone) -> str:
    if missing_quantile is not None:
        return '%'

    return '#'

def num_or_pct_word(missing_tolerance: IntOrNone, missing_quantile: FloatOrNone) -> str:
    if missing_quantile is not None:
        return 'percentage'

    return 'number'

def pct_of_if_necessary(missing_tolerance: IntOrNone, missing_quantile: FloatOrNone) -> str:
    if missing_quantile is not None:
        return '% of'

    return ''

def _t_or_q(missing_tolerance: IntOrNone, missing_quantile: FloatOrNone) -> Union[int, str]:
    if missing_quantile is not None:
        return f'{missing_quantile * 100:g}%'

    return missing_tolerance


import pandas as pd

from datacode.psm.predict import predict_probability_of_treatment
from datacode.psm.match import create_matched_df_using_propensity_scores
from datacode.psm.summarize.main import create_matching_summary_df_dict, create_and_output_matching_summary_latex_table
from datacode.psm.dataprep import treated_and_control_df_from_df
from datacode.psm.names import get_prob_treated_varname
from datacode.psm.typing import StrList, StrListOrNone, DfSummaryTuple, StrOrNone, FloatOrNone, DfDict, TwoDfTuple


def create_matched_df_and_summary(df: pd.DataFrame, treated_var: str, xvars: StrList,
                                  entity_var: str, time_var: str,
                                  fe: StrListOrNone = None,
                                  prob_treated_var_name: StrOrNone = None,
                                  min_matching_pct: FloatOrNone = None,
                                  control_only: bool = False,
                                  num_matches_per_entity: int = 1,
                                  match_with_replacement: bool = False,
                                  control_name: str = 'Control',
                                  treated_name: str = 'Treated', below_text: StrOrNone = None,
                                  outfolder: str = '.', caption: str = 'Propensity Score Matching',
                                  summary_as_table: bool = False
                                  ) -> DfSummaryTuple:
    prob_treated_var_name = get_prob_treated_varname(treated_var, prob_treated_var_name=prob_treated_var_name)
    matched_df, predict_df = create_matched_and_predict_df(
        df,
        treated_var,
        xvars,
        entity_var,
        time_var,
        fe=fe,
        prob_treated_var_name=prob_treated_var_name,
        min_matching_pct=min_matching_pct,
        control_only=False,
        num_matches_per_entity=num_matches_per_entity,
        match_with_replacement=match_with_replacement
    )

    common_args = (
        df,
        matched_df,
        predict_df,
        treated_var,
        xvars,
        entity_var
    )

    common_kwargs = dict(
        fe=fe,
        prob_treated_var=prob_treated_var_name,
        control_name=control_name,
        treated_name=treated_name
    )

    if summary_as_table:
        common_args = common_args + (time_var,)
        common_kwargs.update(dict(
            below_text=below_text,
            outfolder=outfolder,
            caption=caption,
            min_matching_pct=min_matching_pct
        ))
        summary_func = create_and_output_matching_summary_latex_table
    else:
        summary_func = create_matching_summary_df_dict

    df_dict_or_table = summary_func(
        *common_args,
        **common_kwargs
    )

    if control_only:
        _, matched_df = treated_and_control_df_from_df(df, treated_var, entity_var)

    return matched_df, df_dict_or_table


def create_matched_and_predict_df(df: pd.DataFrame, treated_var: str, xvars: StrList,
                                  entity_var: str, time_var: str,
                                  fe: StrListOrNone = None,
                                  prob_treated_var_name: StrOrNone = None,
                                  min_matching_pct: FloatOrNone = None,
                                  control_only: bool = False,
                                  num_matches_per_entity: int = 1,
                                  match_with_replacement: bool = False
                                  ) -> TwoDfTuple:
    prob_treated_var_name = get_prob_treated_varname(treated_var, prob_treated_var_name=prob_treated_var_name)
    predict_df = predict_probability_of_treatment(
        df,
        treated_var,
        xvars,
        entity_var,
        time_var,
        fe=fe
    )

    df[prob_treated_var_name] = predict_df[prob_treated_var_name]

    matched_df = create_matched_df_using_propensity_scores(
        df,
        predict_df,
        treated_var,
        time_var=time_var,
        entity_var=entity_var,
        min_matching_pct=min_matching_pct,
        control_only=control_only,
        num_matches_per_entity=num_matches_per_entity,
        match_with_replacement=match_with_replacement
    )

    return matched_df, predict_df


from datacode.models.variables.variable import Variable

from datacode.models.variables.expression import Expression
from tests.test_variables import VariableTest, VAR1_ARGS, VAR2_ARGS, VAR3_ARGS


class ExpressionTest(VariableTest):
    v = Variable(*VAR1_ARGS)
    v2 = Variable(*VAR2_ARGS)
    v3 = Variable(*VAR3_ARGS)
    expect_add = Expression.from_sympy_expr([v, v2], v.symbol + v2.symbol)
    expect_add_multiple = Expression.from_sympy_expr([v, v2, v3], v.symbol + v2.symbol + v3.symbol)
    expect_subtract = Expression.from_sympy_expr([v, v2], v.symbol - v2.symbol)
    expect_subtract_multiple = Expression.from_sympy_expr([v, v2, v3], v.symbol - v2.symbol - v3.symbol)
    expect_multiply = Expression.from_sympy_expr([v, v2], v.symbol * v2.symbol)
    expect_multiply_multiple = Expression.from_sympy_expr([v, v2, v3], v.symbol * v2.symbol * v3.symbol)
    expect_divide = Expression.from_sympy_expr([v, v2], v.symbol / v2.symbol)
    expect_divide_multiple = Expression.from_sympy_expr([v, v2, v3], v.symbol / v2.symbol / v3.symbol)
    expect_pow = Expression.from_sympy_expr([v, v2], v.symbol ** v2.symbol)
    expect_pow_multiple = Expression.from_sympy_expr([v, v2, v3], v.symbol ** v2.symbol ** v3.symbol)
    add_calc_v = Variable('calc', calculation=expect_add)


class TestCreateFromVariables(ExpressionTest):

    def test_create_from_add(self):
        assert self.v + self.v2 == self.expect_add

    def test_create_from_add_multiple(self):
        assert self.v + self.v2 + self.v3 == self.expect_add_multiple

    def test_create_from_subtract(self):
        assert self.v - self.v2 == self.expect_subtract

    def test_create_from_subtract_multiple(self):
        assert self.v - self.v2 - self.v3 == self.expect_subtract_multiple

    def test_create_from_multiply(self):
        assert self.v * self.v2 == self.expect_multiply

    def test_create_from_multiply_multiple(self):
        assert self.v * self.v2 * self.v3 == self.expect_multiply_multiple

    def test_create_from_divide(self):
        assert self.v / self.v2 == self.expect_divide

    def test_create_from_divide_multiple(self):
        assert self.v / self.v2 / self.v3 == self.expect_divide_multiple

    def test_create_from_pow(self):
        assert self.v ** self.v2 == self.expect_pow

    def test_create_from_pow_multiple(self):
        assert self.v ** self.v2 ** self.v3 == self.expect_pow_multiple




import pandas as pd
from datacode.summarize.subset.outliers.typing import StrList, FloatSequence
from datacode.summarize import format_numbers_to_decimal_places


def bad_df_summary_stats_df(bad_df: pd.DataFrame, outlier_cols: StrList,
                            bad_column_name: str = '_bad_column',
                            quants: FloatSequence=(0.01, 0.1, 0.5, 0.9, 0.99)) -> pd.DataFrame:

    compare_col_df_list = []

    # First add columns from direct functions such as grouped.max()
    for stat in ('min', 'max', 'mean', 'std'):
        this_stat_list = []
        for col in outlier_cols:
            compare_series = bad_df[bad_df[bad_column_name] == col][col]
            compare_value = getattr(compare_series, stat)()
            this_stat_list.append(compare_value)

        compare_col_df_list.append(
            pd.Series(this_stat_list, name=stat, index=outlier_cols).apply(format_numbers_to_decimal_places)
        )

    # Now add columns from grouped.quantile()
    quant_names = []  # track quant names for reordering
    for quant in quants:
        this_stat_list = []
        quant_name = _quant_name(quant)
        quant_names.append(quant_name)
        for col in outlier_cols:
            compare_series = bad_df[bad_df[bad_column_name] == col][col]
            compare_value = compare_series.quantile(quant)
            this_stat_list.append(compare_value)

        compare_col_df_list.append(
            pd.Series(this_stat_list, name=quant_name, index=outlier_cols).apply(format_numbers_to_decimal_places)
        )

    summ_df = pd.concat(compare_col_df_list, axis=1)

    # Column order and final formatting
    summ_df.index.name = 'Variable'
    column_order = ['mean', 'std', 'min'] + quant_names + ['max']
    summ_df = summ_df[column_order]

    return summ_df

def full_df_summary_stats_df(df: pd.DataFrame, outlier_cols: StrList,
                             quants: FloatSequence = (0.01, 0.1, 0.5, 0.9, 0.99)) -> pd.DataFrame:
    summ_df = df[outlier_cols].describe(percentiles=quants).T

    quant_names = [_quant_name(quant) for quant in quants]

    out_columns = ['count', 'mean', 'std', 'min'] + quant_names + ['max']
    summ_df = summ_df[out_columns]
    summ_df['count'] = summ_df['count'].astype(int)

    summ_df = summ_df.applymap(format_numbers_to_decimal_places)

    return summ_df



def _quant_name(quant: float) -> str:
    return f'{quant:.0%}'


from typing import Optional, Sequence

from datacode.models.pipeline.base import DataPipeline
from datacode.models.pipeline.operations.combine import CombineOptions
from datacode.models.types import DataSourcesOrPipelines


class DataCombinationPipeline(DataPipeline):
    """
    A DataPipeline which combines data from two or more DataSources
    """

    def __init__(self, data_sources: DataSourcesOrPipelines,
                 options_list: Sequence[CombineOptions], name: Optional[str] = None):
        super().__init__(
            data_sources,
            options_list,
            name=name
        )

    def execute(self, output: bool = True):
        self._validate()
        super().execute(output=output)

    def _validate(self):
        self._validate_data_sources_combine_options()

    def _validate_data_sources_combine_options(self):
        if len(self.data_sources) - 1 != len(self.operation_options):
            raise ValueError(f'must have one fewer combine options than data sources. Have {len(self.data_sources)} data '
                             f'sources and {len(self.operation_options)} combine options.')



from typing import Optional

from datacode.models.pipeline.operations.generate import GenerationOptions

from datacode.models.pipeline.base import DataPipeline


class DataGeneratorPipeline(DataPipeline):
    """
    A DataPipeline which creates a DataSource without using any other DataSource
    """

    def __init__(self, options: GenerationOptions, name: Optional[str] = None):
        super().__init__(
            [],
            [options],
            name=name
        )



import pandas as pd
from typing import Union, List, Tuple, Dict

StrList = List[str]
ListOrStr = Union[List, str]
DfListListStrTuple = Tuple[pd.DataFrame, List, List, str]
GroupvarNgroupsDict = Dict[str, int]
StrOrInt = Union[int, str]
DictofStrsandStrLists = Dict[str, StrList]
TwoStrTuple = Tuple[str, str]
TwoStrTupleList = List[TwoStrTuple]
StrBoolDict = Dict[str, bool]
TwoStrListTuple = Tuple[StrList, StrList]


import uuid

from graphviz import Digraph

from datacode.graph.base import GraphObject


class Node(GraphObject):

    def __init__(self, name, label='name', **node_kwargs):
        if label == 'name':
            self.label = name
        elif label is None:
            self.label = ''
        else:
            self.label = label

        self.name = name
        self.id = str(uuid.uuid1())
        self.node_kwargs = node_kwargs

    def add_to_graph(self, graph: Digraph):
        graph.node(self.id, label=self.label, **self.node_kwargs)

    def __repr__(self):
        return f'<Node(name={self.name})>'



import datetime
from copy import deepcopy
from typing import Union, Optional, Type

import numpy as np

from datacode.models.dtypes.base import DataType


class DatetimeType(DataType):
    names = ('datetime', 'time', 'datetime64')

    def __init__(self, categorical: bool = False, ordered: bool = False,
                 tz: Optional[Union[str, datetime.timezone]] = None):
        super().__init__(
            datetime.datetime,
            pd_class=np.datetime64,
            categorical=categorical,
            ordered=ordered,
        )
        self.tz = tz
        self.equal_attrs = deepcopy(self.equal_attrs)
        self.equal_attrs.append('tz')
        self.repr_cols = deepcopy(self.repr_cols)
        self.repr_cols.append('tz')

    @classmethod
    def from_str(cls, dtype: str, categorical: bool = False, ordered: bool = False):
        dtype = dtype.lower()
        if dtype not in cls.names:
            raise ValueError(f'Dtype {dtype} does not match valid names for {cls.__name__}: {cls.names}')
        return cls(
            categorical=categorical,
            ordered=ordered
        )

    @property
    def index_arg(self) -> Union[Type, str]:
        return 'datetime64[ns]'



import pandas as pd
import pyexlatex.table as lt

def by_id_count_data_table(df: pd.DataFrame, id_col: str) -> lt.DataTable:
    by_id_count_df = _by_id_count_df(df, id_col)

    by_id_count_dt = lt.DataTable.from_df(
        by_id_count_df,
        extra_header=id_col
    )

    return by_id_count_dt

def _by_id_count_df(df: pd.DataFrame, id_col: str) -> pd.DataFrame:
    by_id_count_df = pd.DataFrame(
        data=[
            (len(df[id_col].dropna().unique()),)
        ],
        columns=['#']
    )

    return by_id_count_df



from typing import Callable, List, Union, Tuple, Optional, Dict, Any, Sequence, TYPE_CHECKING
if TYPE_CHECKING:
    from datacode.models.variables.variable import Variable
    from datacode.models.column.column import Column
import pandas as pd
from functools import partial

from datacode.models.logic.merge.display import display_merge_summary
from datacode.models.pipeline.operations.operation import DataOperation, OperationOptions
from datacode.models.source import DataSource
from datacode.models.logic.merge import left_merge_df
from datacode.summarize import describe_df

StrList = List[str]
StrListOrNone = Union[None, StrList]
TwoDfTuple = Tuple[pd.DataFrame, pd.DataFrame]


class DataMerge(DataOperation):
    num_required_sources = 2
    options: 'MergeOptions'
    result: 'DataSource'

    def __init__(self, data_sources: Sequence[DataSource], merge_options: 'MergeOptions', **result_kwargs):
        self._merged_name = None
        self._merged_type = None
        self._merged_str = None
        super().__init__(
            data_sources,
            merge_options,
            **result_kwargs
        )
        self.output_name = self.merged_name

    def _execute(self):
        print(f'Running merge function {self.merge_str}')
        left_df, right_df = self._get_merge_dfs()
        self.result.df = self.options.merge_function(
            left_df, right_df,
            self.options.on_names,
            **self.options.merge_function_kwargs
        )
        if self.options.post_merge_func is not None:
            self.result.df = self.options.post_merge_func(self.result.df)

        # TODO [#78]: merge source variable combine logic doesn't seem to be working completely correctly
        #
        # Had to put safe=False in merge pipeline output to make it happen
        if 'columns' not in self.result_kwargs:
            left_ds, right_ds = self.data_sources[0], self.data_sources[1]
            load_variables = []
            columns = []
            if left_ds.load_variables:
                for var in left_ds.load_variables:
                    if self.options.left_df_keep_cols is None or var.name in self.options.left_df_keep_cols:
                        load_variables.append(var)
                        columns.append(left_ds.col_for(var))
            if right_ds.load_variables:
                for var in right_ds.load_variables:
                    if self.options.right_df_keep_cols is None or var.name in self.options.right_df_keep_cols:
                        # don't repeat variables and columns. Merge on variables will be repeated,
                        # perhaps even with different transformations, so explictly skip them
                        if var not in load_variables and var.name not in self.options.on_names:
                            load_variables.append(var)
                            columns.append(right_ds.col_for(var))
            self.result.columns = columns
            self.result.load_variables = load_variables

        print(f"""
        {self.data_sources[0].name} obs: {len(left_df)}
        {self.data_sources[1].name} obs: {len(right_df)}
        Merged obs: {len(self.result.df)}
        """)

    def summary(self, *summary_args, summary_method: str=None, summary_function: Callable=None,
                             summary_attr: str=None, **summary_method_kwargs):
        display_merge_summary(
            self,
            *summary_args,
            summary_method=summary_method,
            summary_function=summary_function,
            summary_attr=summary_attr,
            **summary_method_kwargs
        )

    def describe(self):
        display_merge_summary(
            self,
            summary_function=describe_df,
            disp=False # don't display from describe_df as will display from display_merge_summary
        )

    def _get_merge_dfs(self) -> TwoDfTuple:
        left_df = self.data_sources[0].df
        right_df = self.data_sources[1].df

        # Handle pre process funcs
        if self.options.left_df_pre_process_func is not None:
            left_df = self.options.left_df_pre_process_func(left_df)
        if self.options.right_df_pre_process_func is not None:
            right_df = self.options.right_df_pre_process_func(right_df)

        # Handle selecting variables on processed df
        if self.options.left_df_keep_cols is not None:
            left_df = left_df[self.options.left_df_keep_cols]
        if self.options.right_df_keep_cols is not None:
            right_df = right_df[self.options.right_df_keep_cols]

        return left_df, right_df

    @property
    def merged_name(self):
        if self._merged_name is None:
            self._merged_name = f'{self.data_sources[0].name} & {self.data_sources[1].name}'
        return self._merged_name

    @property
    def merge_str(self):
        if self._merged_str is None:
            self._merged_str = f'''
            {self.options.merge_function.__name__}(
                {self.data_sources[0].name},
                {self.data_sources[1].name},
                {self.options.on_names},
                **{self.options.merge_function_kwargs}
            )
            '''
        return self._merged_str

    def __repr__(self):
        return f'<DataMerge(left={self.data_sources[0]}, right={self.data_sources[1]})>'


class LastMergeFinishedException(Exception):
    pass


class MergeOptions(OperationOptions):
    op_class = DataMerge

    def __init__(self, on: Sequence[Union[str, 'Variable', 'Column']], out_path=None, merge_function=left_merge_df,
                 left_df_keep_cols: StrListOrNone=None, right_df_keep_cols: StrListOrNone=None,
                 left_df_pre_process_func: Callable=None, right_df_pre_process_func: Callable=None,
                 left_df_pre_process_kwargs: Optional[Dict[str, Any]] = None,
                 right_df_pre_process_kwargs: Optional[Dict[str, Any]] = None,
                 post_merge_func: Callable = None, post_merge_func_kwargs: Optional[Dict[str, Any]] = None,
                 allow_modifying_result: bool = True, result_kwargs: Optional[Dict[str, Any]] = None,
                 **merge_function_kwargs):
        """

        passed args to merge func will be (
            left_df_pre_process_func(left_df, **left_df_pre_process_kwargs)[left_df_keep_cols],
            right_df_pre_process_func(right_df, **right_df_pre_process_kwargs)[right_df_keep_cols],
            *merge_function_kwargs,
            **merge_function_kwargs
        )

        if left_df_keep_cols is None, will instead pass left_df_pre_process_func(left_df, **left_df_pre_process_kwargs).
        If right_df_keep_cols is None, will instead pass right_df_pre_process_func(right_df, **right_df_pre_process_kwargs).
        If left_df_pre_process_func is None, will instead pass left_df or left_df[left_df_keep_cols] depending
            on whether left_df_keep_cols was passed. Similar behavior for right.

        Args:
            *merge_function_args:
            out_path:
            merge_function:
            left_df_keep_cols:
            right_df_keep_cols:
            post_merge_func: function to be called on data after merge
            post_merge_func_kwargs: kwargs to be passed to post_merge_func
            **merge_function_kwargs:

        :param allow_modifying_result: When DataSources are directly linked to pipelines, loading
            source from pipeline can cause modifications in the pipeline's result source. Set to False
            to ensure it won't be modified (but uses more memory). Setting to False should only be needed
            if multiple sources load from the same pipeline in one session
        """
        from datacode.models.variables.variable import Variable
        from datacode.models.column.column import Column

        if left_df_pre_process_kwargs is None:
            left_df_pre_process_kwargs = {}

        if right_df_pre_process_kwargs is None:
            right_df_pre_process_kwargs = {}

        if post_merge_func_kwargs is None:
            post_merge_func_kwargs = {}

        if left_df_pre_process_func is None:
            left_df_pre_process_func = lambda x: x

        if right_df_pre_process_func is None:
            right_df_pre_process_func = lambda x: x

        if post_merge_func is None:
            post_merge_func = lambda x: x

        on_names = []
        for item in on:
            if isinstance(item, str):
                on_names.append(item)
            elif isinstance(item, Variable):
                on_names.append(item.name)
            elif isinstance(item, Column):
                on_names.append(item.variable.name)

        self.on = on
        self.on_names = on_names
        self.out_path = out_path
        self.merge_function = merge_function
        self.merge_function_kwargs = merge_function_kwargs
        self.left_df_keep_cols = left_df_keep_cols
        self.right_df_keep_cols = right_df_keep_cols
        self.left_df_pre_process_func = partial(left_df_pre_process_func, **left_df_pre_process_kwargs)
        self.right_df_pre_process_func = partial(right_df_pre_process_func, **right_df_pre_process_kwargs)
        self.post_merge_func = partial(post_merge_func, **post_merge_func_kwargs)
        self.allow_modifying_result = allow_modifying_result
        self.result_kwargs = result_kwargs

    def __repr__(self):
        return f'<DataMerge(on_names={self.on_names}, merge_function={self.merge_function.__name__}, ' \
               f'kwargs={self.merge_function_kwargs})>'

    def update(self, **kwargs):
        self.merge_function_kwargs.update(**kwargs)


import pandas as pd

from datacode.portfolio.sort import sort_portfolio_index, sort_portfolio_columns

def sort_and_apply_formatting_to_df(counts: pd.DataFrame,
                                  sort_cols_as_numeric: bool=True, sort_rows_as_numeric: bool=True,
                                  sort_cols_as_portvar: bool=False, sort_rows_as_portvar: bool=False,
                                  format_str: str='.0f'
                                  ) -> pd.DataFrame:

    if sort_cols_as_numeric:
        counts.columns = [_to_int_if_possible(col) for col in counts.columns]
        counts.sort_index(axis=1, inplace=True)

    if sort_rows_as_numeric:
        counts.index = [_to_int_if_possible(col) for col in counts.index]
        counts.sort_index(inplace=True)

    if sort_cols_as_portvar:
        assert not sort_cols_as_numeric
        counts = sort_portfolio_columns(counts)

    if sort_rows_as_portvar:
        assert not sort_rows_as_numeric
        counts = sort_portfolio_index(counts)

    return counts.applymap(lambda x: f'{x:{format_str}}')

def _to_int_if_possible(s):
    try:
        return int(float(s))
    except ValueError:
        return s


from .main import Pipeline, LogicalPipeline, DataMergePipeline, PipelineOptions
from .datasets import DataSource, DataSubject
from .bars import MatchComparisonBarGraph
from datacode.compareids.models.interface import MatchComparisonBarData
from .steps import Step, MergeStep, Process


from typing import Tuple

from datacode.models.variables import Variable, VariableCollection
from datacode import Transform, AppliedTransform
from datacode.models.symbols import Symbol
from tests.input_funcs import lag

VAR1_ARGS = ('stuff', 'My Stuff')
VAR2_ARGS = ('thing_me', 'My Thing')
VAR3_ARGS = ('other', 'My Other')
VC_NAME = 'my_collection'
TRANSFORMS = [
    Transform('port', name_func=lambda x: f'{x} Port Var'),
    Transform('lag', name_func=lag),
]


class VariableTest:

    def create_variables(self, **kwargs) -> Tuple[Variable, Variable]:
        config_dict = dict(

        )
        config_dict.update(**kwargs)
        v = Variable(*VAR1_ARGS, **config_dict)
        v2 = Variable(*VAR2_ARGS, **config_dict)
        return v, v2

    def create_variable_collection(self, **kwargs) -> Tuple[VariableCollection, Variable, Variable]:
        config_dict = dict(
            name=VC_NAME
        )
        config_dict.update(**kwargs)
        v, v2 = self.create_variables()
        vc = VariableCollection(v, v2, **config_dict)
        return vc, v, v2


class TestVariable(VariableTest):

    def test_create_variables(self):
        v, v2 = self.create_variables()

        assert v.name == VAR1_ARGS[1]
        assert v.key == VAR1_ARGS[0]
        assert v.to_tuple() == VAR1_ARGS
        assert v.symbol == Symbol(r'\text{Stuff}')
        assert v2.name == VAR2_ARGS[1]
        assert v2.key == VAR2_ARGS[0]
        assert v2.symbol == Symbol('TM')
        assert v2.to_tuple() == VAR2_ARGS

    def test_set_description(self):
        desc = 'my desc'
        v, v2 = self.create_variables(description=desc)
        assert v.description == desc
        assert v2.description == desc


class TestVariableCollection(VariableTest):

    def test_create_variable_collection(self):
        vc, v, v2 = self.create_variable_collection()

        assert vc.name == VC_NAME
        assert vc.stuff == v
        assert vc.thing_me == v2

    def test_create_variable_collection_with_transforms(self):
        vc, v, v2 = self.create_variable_collection(
            transforms=TRANSFORMS
        )

        assert vc.name == VC_NAME
        assert vc.stuff == v
        assert vc.thing_me == v2
        assert vc.stuff.port().name == VAR1_ARGS[1] + ' Port Var'
        assert vc.stuff.port().port().name == VAR1_ARGS[1] + ' Port Var Port Var'
        assert str(vc.stuff.port().symbol) == r'\text{Stuff} Port Var'
        assert str(vc.stuff.port().port().symbol) == r'\text{Stuff} Port Var Port Var'
        assert vc.thing_me.port().name == VAR2_ARGS[1] + ' Port Var'
        assert str(vc.thing_me.port().symbol) == 'TM Port Var'
        assert vc.stuff.lag(1).name == VAR1_ARGS[1] + '_{t - 1}'
        assert str(vc.stuff.lag(1).symbol) == r'\text{Stuff}_{t - 1}'
        assert vc.thing_me.lag(1).name == VAR2_ARGS[1] + '_{t - 1}'
        assert str(vc.thing_me.lag(1).symbol) == 'TM_{t - 1}'

    def test_create_variable_collection_with_default_attr(self):
        vc, v, v2 = self.create_variable_collection(
            default_attr='name'
        )

        assert vc.name == VC_NAME
        assert vc.stuff == VAR1_ARGS[1]
        assert vc.thing_me == VAR2_ARGS[1]

    def test_create_variable_collection_with_default_transform_no_args(self):
        vc, v, v2 = self.create_variable_collection(
            transforms=TRANSFORMS,
            default_transforms=[AppliedTransform.from_transform(TRANSFORMS[0])],
        )

        assert vc.name == VC_NAME
        assert vc.stuff.name == VAR1_ARGS[1] + ' Port Var'
        assert str(vc.stuff.symbol) == r'\text{Stuff} Port Var'
        assert vc.thing_me.name == VAR2_ARGS[1] + ' Port Var'
        assert str(vc.thing_me.symbol) == 'TM Port Var'

    def test_create_variable_collection_with_default_transform_with_args(self):
        vc, v, v2 = self.create_variable_collection(
            transforms=TRANSFORMS,
            default_transforms=[AppliedTransform.from_transform(TRANSFORMS[1], 5)],
        )

        assert vc.name == VC_NAME
        assert vc.stuff.name == VAR1_ARGS[1] + '_{t - 5}'
        assert str(vc.stuff.symbol) == r'\text{Stuff}_{t - 5}'
        assert vc.thing_me.name == VAR2_ARGS[1] + '_{t - 5}'
        assert str(vc.thing_me.symbol) == 'TM_{t - 5}'

    def test_create_variable_collection_with_multiple_default_transforms(self):
        vc, v, v2 = self.create_variable_collection(
            transforms=TRANSFORMS,
            default_transforms=[
                AppliedTransform.from_transform(TRANSFORMS[0]),
                AppliedTransform.from_transform(TRANSFORMS[1], 5)
            ],
        )

        assert vc.name == VC_NAME
        assert vc.stuff.name == VAR1_ARGS[1] + ' Port Var_{t - 5}'
        assert str(vc.stuff.symbol) == r'\text{Stuff} Port Var_{t - 5}'
        assert vc.thing_me.name == VAR2_ARGS[1] + ' Port Var_{t - 5}'
        assert str(vc.thing_me.symbol) == 'TM Port Var_{t - 5}'

    def test_create_variable_collection_with_default_attr_and_default_transform(self):
        vc, v, v2 = self.create_variable_collection(
            transforms=TRANSFORMS,
            default_transforms=[AppliedTransform.from_transform(TRANSFORMS[0])],
            default_attr='name'
        )

        assert vc.name == VC_NAME
        assert vc.stuff == VAR1_ARGS[1] + ' Port Var'
        assert vc.thing_me == VAR2_ARGS[1] + ' Port Var'

    def test_create_multiple_variable_collections_no_overlap(self):
        vc, v, v2 = self.create_variable_collection()
        vc2 = VariableCollection(
            v,
            v2,
            name=VC_NAME,
            transforms=TRANSFORMS,
            default_transforms=[AppliedTransform.from_transform(TRANSFORMS[0])],
        )

        assert vc.name == VC_NAME
        assert vc2.name == VC_NAME
        assert vc.stuff == v
        assert vc.thing_me == v2
        assert not vc.stuff.applied_transforms
        assert not vc.thing_me.applied_transforms
        assert vc2.stuff.name == VAR1_ARGS[1] + ' Port Var'
        assert vc2.thing_me.name == VAR2_ARGS[1] + ' Port Var'
        assert vc2.stuff.applied_transforms
        assert vc2.thing_me.applied_transforms




"""
Data pipelines for humans. Work with variables, sources, and pipelines rather than raw data. Also includes high-level tools to analyze, summarize, and transform data.
"""
from datacode.models.source import DataSource
from datacode.models.loader import DataLoader
from datacode.models.outputter import DataOutputter, DataOutputNotSafeException
from datacode.models.index import Index
from datacode.models.column.column import Column
from datacode.models.column.index import ColumnIndex
from datacode.models.pipeline.merge import DataMergePipeline
from datacode.models.pipeline.generate import DataGeneratorPipeline
from datacode.models.pipeline.analysis import DataAnalysisPipeline
from datacode.models.pipeline.transform import DataTransformationPipeline
from datacode.models.pipeline.combine import DataCombinationPipeline
from datacode.models.pipeline.operations.merge import MergeOptions
from datacode.models.pipeline.operations.analysis import AnalysisOptions
from datacode.models.pipeline.operations.transform import TransformOptions
from datacode.models.pipeline.operations.generate import GenerationOptions
from datacode.models.pipeline.operations.combine import CombineOptions
from datacode.models.analysis import AnalysisResult
from datacode.models.variables.variable import Variable
from datacode.models.transform.transform import Transform
from datacode.models.transform.applied import AppliedTransform
from datacode.models.transform.source import SourceTransform
from datacode.models.transform.specific import DEFAULT_TRANSFORMS
from datacode.models.variables.collection import VariableCollection
from datacode.models.variables.expression import Expression
from datacode.summarize import describe_df

from datacode.models.dtypes.base import DataType
from datacode.models.dtypes.bool_type import BooleanType
from datacode.models.dtypes.datetime_type import DatetimeType
from datacode.models.dtypes.float_type import FloatType
from datacode.models.dtypes.int_type import IntType
from datacode.models.dtypes.obj_type import ObjectType
from datacode.models.dtypes.str_type import StringType
from datacode.models.dtypes.timedelta_type import TimedeltaType

from datacode.models.transform.plugin import register_transforms

from datacode.sem import SEM, SEMSummary
from datacode.summarize.coverage.main import variables_drop_obs_doc



import pandas as pd
from datacode.psm.score import comparison_arrays_from_treated_and_control_df, comparison_score


def yield_matched_scores(treated_df: pd.DataFrame, control_df: pd.DataFrame, match_dict: dict,
                         prob_treated_var: str, time_var: str = 'Date',
                         entity_var: str = 'id') -> pd.DataFrame:
    for treated_id, control_id in match_dict.items():
        arrays = comparison_arrays_from_treated_and_control_df(
            treated_df[treated_df[entity_var] == treated_id],
            control_df[control_df[entity_var] == control_id],
            prob_treated_var,
            time_var=time_var
        )
        df = pd.DataFrame([
            pd.Series(arr, name=name) for name, arr in
            {f'Treated: {treated_id}': arrays[0], f'Control: {control_id}': arrays[1]}.items()
        ]).T
        score = comparison_score(arrays)
        print(f'Score: {score}')
        yield df



from datacode.psm.typing import DfDict, StrOrNone, FloatOrNone
import pyexlatex.table as lt

def matching_latex_table_from_df_dict(df_dict: DfDict, entity_var: str, time_var: str, below_text: StrOrNone = None,
                                      caption: str = 'Propensity Score Matching',
                                      min_matching_pct: FloatOrNone = None) -> lt.Table:
    model_dt = lt.DataTable.from_df(
        df_dict['model'],
        include_index=True,
        extra_header='Coefficients and Standard Errors',
    )

    predict_dt = lt.DataTable.from_df(
        df_dict['predict'],
        include_index=True,
        extra_header='Prediction Accuracy',
    )

    logit_panel = lt.Panel.from_data_tables(
        [model_dt, predict_dt],
        shape=(1, 2),
        name='Logistic Regression Results'
    )

    match_panel = lt.Panel.from_df(
        df_dict['match'],
        include_index=True,
        name='Summary Statistics by Sample'
    )

    if below_text is None:
        below_text = """
        Propensity Score Matching (PSM) is used to create a sample of non-treated which is similar
        to the treated firms across observable characteristics.
        """

    if min_matching_pct is not None:
        below_text += ' ' + f"""
        When finding matches, only those {entity_var}s with {min_matching_pct * 100:.3g}% of overlapping
        {time_var} observations were used.
        """.strip()

    table = lt.Table.from_panel_list(
        [logit_panel, match_panel],
        below_text=below_text,
        caption=caption
    )

    return table


from typing import Tuple, Optional

from graphviz import Digraph

from datacode.graph.base import GraphObject
from datacode.graph.node import Node
from datacode.graph.subgraph import Subgraph


class Edge(GraphObject):

    def __init__(self, start: Node, end: Node, for_subgraphs: Optional[Tuple[Subgraph, Subgraph]] = None, **edge_kwargs):
        self.start = start
        self.end = end
        if for_subgraphs is not None:
            edge_kwargs.update({'ltail': 'cluster' + for_subgraphs[0].name, 'lhead': 'cluster' + for_subgraphs[1].name})
        self.edge_kwargs = edge_kwargs

    def add_to_graph(self, graph: Digraph):
        graph.edge(self.start.id, self.end.id, **self.edge_kwargs)

    def __repr__(self):
        return f'<Edge(start={self.start}, end={self.end}, edge_kwargs: {self.edge_kwargs})>'



import operator
from copy import deepcopy
from typing import Sequence, Optional, Union, Callable

from mixins import ReprMixin

from datacode.models.dtypes.base import DataType
from datacode.models.dtypes.convert import convert_str_to_data_type_if_necessary
from datacode.models.variables.expression import Expression
from datacode.models.transform.transform import Transform
from datacode.models.transform.applied import AppliedTransform
from datacode.models.symbols import Symbol, var_key_to_symbol_str, to_symbol_if_necessary


class Variable(ReprMixin):
    repr_cols = ['key', 'name', 'applied_transforms']

    def __init__(self, key: str, name: Optional[str]=None, symbol: Optional[Union[str, Symbol]] = None,
                 dtype: Optional[Union[str, DataType]] = None,
                 available_transforms: Optional[Sequence[Transform]] = None,
                 applied_transforms: Optional[Sequence[AppliedTransform]] = None,
                 description: str = '', calculation: Optional[Expression] = None):
        if symbol is None:
            symbol = var_key_to_symbol_str(key)
        symbol = to_symbol_if_necessary(symbol)

        if available_transforms is None:
            available_transforms = []

        if applied_transforms is None:
            applied_transforms = []

        dtype = convert_str_to_data_type_if_necessary(dtype)

        self.key = key
        self.dtype = dtype
        self.symbol = symbol
        self.available_transforms = available_transforms
        self.applied_transforms = applied_transforms
        self.description = description
        self.calculation = calculation

        if name is None:
            name = _from_var_name_to_display_name(key)
        self.name = name
        self._orig_name = name
        self._orig_symbol = symbol
        self._update_from_transforms()

    def __eq__(self, other):
        compare_attrs = ('key', 'applied_transforms')
        # If any comparison attribute is missing in the other object, not equal
        if any([not hasattr(other, attr) for attr in compare_attrs]):
            return False
        # If all compare attributes are equal, objects are equal
        return all([getattr(self, attr) == getattr(other, attr) for attr in compare_attrs])

    def __hash__(self):
        applied_transform_hash = hash(tuple([hash(transform) for transform in self.applied_transforms]))
        return hash((self.key, applied_transform_hash))

    def __add__(self, other) -> Expression:
        return self._create_expression_from_other_and_operator(other, operator.add, 'add')

    def __sub__(self, other):
        return self._create_expression_from_other_and_operator(other, operator.sub, 'subtract', preposition='from')

    def __mul__(self, other):
        return self._create_expression_from_other_and_operator(other, operator.mul, 'multiply', preposition='by')

    def __truediv__(self, other):
        return self._create_expression_from_other_and_operator(other, operator.truediv, 'divide', preposition='by')

    def __pow__(self, other):
        return self._create_expression_from_other_and_operator(other, operator.pow, 'exponentiate', preposition='by')

    @property
    def unique_key(self) -> str:
        key = self.key
        if not self.applied_transforms:
            return key

        transform_keys = [transform.key for transform in self.applied_transforms]
        full_key = '_'.join([key] + transform_keys)
        return full_key

    def _create_expression_from_other_and_operator(self, other, op_func: Callable, operator_name: str,
                                                   preposition: str = 'to') -> 'Expression':
        if isinstance(other, Variable):
            sympy_expr = op_func(self.symbol, other.symbol)
            expr = Expression.from_sympy_expr([self, other], sympy_expr)
            return expr

        if isinstance(other, Expression):
            if other.expr is None:
                raise ValueError(f'cannot {operator_name} expression which does not have .expr, got {other}')
            sympy_expr = op_func(self.symbol, other.expr)
            expr = Expression.from_sympy_expr([self, *other.variables], sympy_expr)
            return expr

        raise ValueError(f'Cannot {operator_name} {other} of type {type(other)} {preposition} {self}, '
                         f'must be Variable or Expression')


    def to_tuple(self):
        return self.key, self.name

    @classmethod
    def from_display_name(cls, display_name):
        name = _from_display_name_to_var_name(display_name)
        return cls(name, display_name)

    def _add_transform(self, transform: Transform):
        self.available_transforms.append(transform)
        self._add_transform_attr(transform)
        self._set_name_and_symbol_by_transforms()

    def _add_applied_transform(self, transform: Transform):
        self.applied_transforms.append(transform)
        self._add_transform(transform)

    def _update_from_transforms(self):
        """
        Adds attributes for current available_transforms

        Notes:
            Does not remove any previous attributes for previous available_transforms
        """

        # Add attributes for available transforms
        for transform in self.available_transforms:
            self._add_transform_attr(transform)

        # Apply name changes from applied transforms
        self._set_name_and_symbol_by_transforms()

    def _set_name_and_symbol_by_transforms(self):
        """
        Apply name changes from applied transforms
        """
        name = self._orig_name
        sym = self._orig_symbol
        for transform in self.applied_transforms:
            if transform.name_func is not None:
                name = transform.name_func(name)
            if transform.symbol_func is not None:
                sym = transform.symbol_func(sym)
        self.name = name
        self.symbol = sym

    def _add_transform_attr(self, transform: Transform):
        def transform_func(*args, **kwargs):
            transform_value = deepcopy(self)
            applied_transform = AppliedTransform.from_transform(transform, *args, **kwargs)
            transform_value.applied_transforms.append(applied_transform)
            transform_value._update_from_transforms()
            transform_value._set_name_and_symbol_by_transforms()
            return transform_value
        setattr(self, transform.key, transform_func)


def _from_var_name_to_display_name(var_name):
    return ' '.join([word for word in var_name.split('_')]).title()


def _from_display_name_to_var_name(display_name):
    return '_'.join([word for word in display_name.split(' ')]).lower()



import uuid
from typing import Optional, Callable, Any, TYPE_CHECKING
if TYPE_CHECKING:
    from datacode.models.variables.variable import Variable
    from datacode.models.column.column import Column
    from datacode.models.source import DataSource

from datacode.models.transform.transform import Transform
from datacode.models.logic.partial import partial
from datacode.models.variables.typing import StrFunc, ValueFunc, SymbolFunc


class AppliedTransform(Transform):
    """
    Works like Transform but allows passing of arg and kwargs in advance, similar to functools.partial
    """
    repr_cols = ['key', 'name_func', 'data_func', 'symbol_func', 'data_func_target', 'args', 'kwargs']

    def __init__(self, key: str, *args, name_func: StrFunc = None, data_func: ValueFunc = None,
                 symbol_func: SymbolFunc = None,
                 data_func_target: Optional[str] = None, **kwargs):
        super().__init__(
            key,
            name_func=name_func,
            symbol_func=symbol_func,
            data_func=data_func,
            data_func_target=data_func_target
        )
        self.args = args
        self.kwargs = kwargs

        if self.name_func is not None:
            self.name_func = partial(self.name_func, ..., *args, **kwargs)
        if self.symbol_func is not None:
            self.symbol_func = partial(self.symbol_func, ..., *args, **kwargs)
        if self.data_func is not None:
            self.data_func = partial(self.data_func, ..., *args, **kwargs)

    @classmethod
    def from_transform(cls, transform: Transform, *args, **kwargs):
        obj = cls(
            transform.key,
            *args,
            name_func=transform.name_func,
            symbol_func=transform.symbol_func,
            data_func=transform.data_func,
            data_func_target=transform.data_func_target,
            **kwargs)
        return obj

    def __eq__(self, other) -> bool:
        same = super().__eq__(other)
        if not same:
            return False
        try:
            same = same and self.args == other.args
            same = same and self.kwargs == other.kwargs
            return same
        except AttributeError:
            return False

    # TODO [#69]: should not be necessary to explicitly hash AppliedTransform
    #
    # Already hashing `Transform` base class. But was getting unhashable type: 'AppliedTransform'
    # even with that.
    def __hash__(self):
        hash_attrs = ['key', 'data_func_target']
        hash_items = tuple([getattr(self, attr) for attr in hash_attrs])
        return hash(hash_items)

    @classmethod
    def from_func(cls, func: Callable[['Column', 'Variable', 'DataSource', Any], 'DataSource'],
                  key: Optional[str] = None,
                  data_func_target: str = 'source', **func_kwargs):
        if key is None:
            key = str(uuid.uuid4())
        return cls(
            key,
            data_func=func,
            data_func_target=data_func_target,
            **func_kwargs
        )


from typing import Optional, Union

from mixins import ReprMixin

from datacode.models.dtypes.base import DataType
from datacode.models.dtypes.convert import convert_str_to_data_type_if_necessary


class Index(ReprMixin):
    repr_cols = ['key', 'dtype']

    def __init__(self, key: str, dtype: Optional[Union[str, DataType]] = None):
        dtype = convert_str_to_data_type_if_necessary(dtype)

        self.key = key
        self.dtype = dtype

    def __eq__(self, other):
        if not isinstance(other, Index):
            return False

        return all([
            self.key == other.key,
            self.dtype == other.dtype
        ])



import uuid
from copy import deepcopy
from functools import partial
from typing import TYPE_CHECKING, Callable, Optional, Sequence


if TYPE_CHECKING:
    from datacode.models.source import DataSource
    from datacode.models.variables.variable import Variable
    from datacode.models.column.column import Column

from datacode.models.transform.transform import Transform
from datacode.models.variables.typing import StrFunc, ValueFunc, SymbolFunc
import datacode.hooks as hooks


class SourceTransform(Transform):
    """
    Tracks and applies changes to an entire data source for data, name, and symbol together
    """
    repr_cols = ['key', 'name_func', 'data_func', 'symbol_func', 'subset']

    def __init__(self, key: str, name_func: StrFunc = None, data_func: ValueFunc = None,
                 symbol_func: SymbolFunc = None, subset: Sequence['Variable'] = None):
        super().__init__(
            key,
            name_func=name_func,
            data_func=data_func,
            symbol_func=symbol_func,
            data_func_target='source'
        )
        self.subset = subset

    def apply(self, source: 'DataSource', preserve_original: bool = True) -> 'DataSource':
        """
        Applies transformation to data source

        :param source:
        :param preserve_original: True to copy the source before applying transformations. False will decrease
        memory usage but will cause the original source to be partially modified
        :return:
        """
        source = hooks.on_begin_apply_source_transform(self, source)

        from datacode.models.source import NoColumnForVariableException
        if preserve_original:
            source = deepcopy(source)
        else:
            # Even when not preserving original, don't want to modify original variables or columns
            # as they may be used in other sources

            # TODO [#64]: Preserving variables in transform apply to source inplace not working
            #
            # This code is supposed to prevent that but is not working as expected.
            # The original variables are still being modified. The problem occurs with both
            # SourceTransform.apply and Transform.apply_to_source. A test has been added which
            # catches this issue in test_lags_as_source_transform_with_subset but it has been
            # commented out for now.
            source.load_variables = deepcopy(source.load_variables)
            source.columns = deepcopy(source.columns)

        # Call transformation on source data
        if self.data_func is not None:
            source = self.data_func(source)

        if self.subset is None:
            subset = source.load_variables
        else:
            subset = deepcopy(self.subset)
        if subset is None:
            subset = []

        # Collect necessary renames
        rename_dict = {}
        for selected_var in subset:
            try:
                col = source.col_for(variable=selected_var)
            except NoColumnForVariableException:
                # Must have removed this column in the transformation
                continue
            var = col.variable  # Don't use variable directly besides key as may be different instance

            # Update variable
            orig_name = var.name
            col._add_applied_transform(self)
            new_name = var.name

            # Update column name
            if orig_name != new_name:
                rename_dict[orig_name] = new_name

        if rename_dict:
            source.df.rename(columns=rename_dict, inplace=True)

        source = hooks.on_end_apply_source_transform(self, source)
        return source

    @classmethod
    def from_func(cls, func: Callable[['DataSource'], 'DataSource'] = None, key: Optional[str] = None,
                  subset: Sequence['Variable'] = None):
        if key is None:
            key = str(uuid.uuid4())
        return cls(
            key,
            data_func=func,
            subset=subset
        )

    @classmethod
    def from_transform(cls, transform: Transform, subset: Sequence['Variable'] = None):
        # Transform is by variable, need to create new data function which
        # accepts source and applies to subset of variables one by one
        def data_func(subset: Optional[Sequence[str]], source: 'DataSource', **kwargs) -> 'DataSource':
            if subset is None:
                subset = source.load_variables

            for variable in subset:
                col = source.col_for(variable=variable)
                source = transform.data_func(col, variable, source, **kwargs)

            return source

        data_func = partial(data_func, subset)

        return cls(
            transform.key,
            name_func=transform.name_func,
            data_func=data_func,
            symbol_func=transform.symbol_func,
            subset=subset,
        )



from .steps import Process
from .interface import PipelineOptions
from .flowchart import _get_digraph_if_true_else_graph

class Pipeline:

    def __init__(self, process: Process, options: [PipelineOptions, None] = None, name=None):
        self.process = process
        if options is None:
            options = PipelineOptions() #initialize with defaults
        if name is None:
            name = 'default graph'
        self.options = options
        self.name = name

    def build_graph(self, **graph_kwargs):
        subgraphs = self.process.to_subgraphs(self.options)
        edges = self.process.to_edges(self.options)
        graph_elements = subgraphs + edges
        if 'graph_attr' in graph_kwargs:
            graph_kwargs['graph_attr'].update({'compound': 'true'})
        else:
            graph_kwargs['graph_attr'] = {'compound': 'true'}

        Graph = _get_digraph_if_true_else_graph(self.options.digraph)
        graph = Graph(self.name, **graph_kwargs)
        [elem.add_to_graph(graph) for elem in graph_elements]

        self.graph = graph
        return graph


class LogicalPipeline(Pipeline):

    def __init__(self, process: Process, options: [PipelineOptions, None] = None):
        if options is None:
            options = PipelineOptions(data_sources=False) #initialize with defaults
        else:
            options.data_sources = False
        super().__init__(process, options)


class DataMergePipeline(Pipeline):

    def __init__(self, process: Process, options: [PipelineOptions, None] = None):
        if options is None:
            options = PipelineOptions(data_sources=True) #initialize with defaults
        else:
            options.data_sources = True
        super().__init__(process, options)



import itertools

def compare_identifiers(df_tup_list, id_var = 'CQID'):
    """
    df_tup_list is in the format
    df_tup_list = [
    ('Financials', df),
    ('Government Ownership', gdf),
    ]
    """

    id_lists = [df[id_var].unique().tolist() for key, df in df_tup_list]

    for i, j in itertools.combinations([i for i in range(len(df_tup_list))], 2):

        output_str = _compare_idlists_and_output_summary(
            id_lists[i],
            id_lists[j],
            df_tup_list[i][0],
            df_tup_list[j][0]
        )
        print(output_str)

    print(', '.join([_separate_summary(tup[0], id_lists[i]) for i, tup in enumerate(df_tup_list)]))

def _compare_idlists_and_output_summary(own_ids, other_ids, own_name, other_name):
    output_str = ''
    output_str += _compare_idlists_and_output_summary_one_way(
        own_ids,
        other_ids,
        own_name,
        other_name
    )
    output_str += _compare_idlists_and_output_summary_one_way(
        other_ids,
        own_ids,
        other_name,
        own_name
    )

    return output_str


def _compare_idlists_and_output_summary_one_way(own_ids, other_ids, own_name, other_name):

    own_ids_in_other = len([item for item in own_ids if item in other_ids])

    return f'{own_ids_in_other} ids from {own_name} in {other_name}.\n'

def _separate_summary(name, ids):
    return f'{len(ids)} ids in {name}'


import re
from typing import Optional

DTYPE_PATTERN = re.compile(r'[a-z_]+(\d+)?')

def get_bit_from_dtype(dtype: str) -> Optional[int]:
    match = DTYPE_PATTERN.match(dtype)
    if not match:
        raise ValueError(f'could not parse {dtype} as dtype')
    str_or_none = match.group(1)
    if str_or_none:
        # Got an int as a str
        return int(str_or_none)

    # Got None, just return it
    if str_or_none is None:
        return str_or_none

    raise ValueError


import datetime
from typing import Callable, Optional, Any, Sequence, Dict

from datacode.models.logic.combine import CombineFunction, combine_sources
from datacode.models.pipeline.operations.operation import DataOperation, OperationOptions
from datacode.models.source import DataSource


class CombineOperation(DataOperation):
    """
    Data operation that combines data from two DataSources
    """
    num_required_sources = 2
    options: 'CombineOptions'
    result: 'DataSource'

    def __init__(self, data_sources: Sequence[DataSource], options: 'CombineOptions', **result_kwargs):
        super().__init__(
            data_sources,
            options,
            **result_kwargs
        )
        possible_last_modified = [source.last_modified for source in self.data_sources if source.last_modified]
        if possible_last_modified:
            self.options.last_modified = max(possible_last_modified)
        else:
            self.options.last_modified = None

    def _execute(self):
        ds = self.options.func(self.data_sources, **self.options.func_kwargs)
        self.result.update_from_source(ds)
        return self.result

    def summary(self, *summary_args, summary_method: str=None, summary_function: Callable=None,
                             summary_attr: str=None, **summary_method_kwargs):
        # TODO [#73]: better summary for DataCombinationPipeline
        print(f'Combining {self.data_sources[0]} with {self.data_sources[1]} with '
              f'options {self.options}')

    def describe(self):
        return self.summary()

    def __repr__(self):
        return f'<CombinationOperation(options={self.options})>'


class CombineOptions(OperationOptions):
    """
    Class for options passed to AnalysisOperations
    """
    op_class = CombineOperation

    def __init__(self, func: CombineFunction = combine_sources,
                 out_path: Optional[str] = None,
                 result_kwargs: Optional[Dict[str, Any]] = None,
                 **func_kwargs):
        """

        :param func: function which combines the DataSources
        :param out_path: location for generated DataSource
        :param func_kwargs: Keyword arguments to pass to the function which generates the DataSource
        """
        self.func = func
        self.func_kwargs = func_kwargs
        self.out_path = out_path
        self.result_kwargs = result_kwargs



from typing import Sequence

from mixins import ReprMixin

from datacode.models.index import Index
from datacode.models.variables import Variable


class ColumnIndex(ReprMixin):
    repr_cols = ['index', 'variables']

    def __init__(self, index: Index, variables: Sequence[Variable]):
        self.index = index
        self.variables = variables

    def __eq__(self, other):
        if not isinstance(other, ColumnIndex):
            return False

        return all([
            self.index == other.index,
            sorted(self.variables) == sorted(other.variables)
        ])



from typing import Callable


def get_summary_of_df(df, *summary_args, summary_method: str=None, summary_function: Callable=None,
                      summary_attr: str=None, **summary_method_kwargs):
    try:
        if summary_method is not None:
            return getattr(df, summary_method)(*summary_args, **summary_method_kwargs)
        if summary_function is not None:
            return summary_function(df, *summary_args, **summary_method_kwargs)
        if summary_attr is not None:
            return getattr(df, summary_attr)
    except Exception as e:
        # keep going, but save exception as result
        return e


from typing import Sequence

import pandas as pd
import sklearn.preprocessing as sklp

from datacode.models.variables.variable import Variable


def assemble_model_data(df: pd.DataFrame, variables: Sequence[Variable], scale: bool = True,
                        robust_scale: bool = False, **scale_kwargs) -> pd.DataFrame:
    _validate_scale(scale, robust_scale)
    mod_df = _assemble_model_data(df, variables)
    if scale:
        sklp.scale(mod_df, copy=False, **scale_kwargs)
    if robust_scale:
        sklp.robust_scale(mod_df, copy=False, **scale_kwargs)

    return mod_df


def _assemble_model_data(df: pd.DataFrame, variables: Sequence[Variable]) -> pd.DataFrame:
    var_names = [var.name for var in variables]
    rename_dict = {var.name: var.unique_key for var in variables}
    return df[var_names].rename(columns=rename_dict).reset_index(drop=True)


def _validate_scale(scale: bool, robust_scale: bool):
    if scale and robust_scale:
        raise ValueError('must select one or neither of scale or robust_scale, not both')



import pandas as pd

from datacode.summarize.subset.missing.detail.reformat.towide import _long_counts_to_wide_df
from datacode.summarize.subset.missing.detail.reformat.numformat import sort_and_apply_formatting_to_df
from datacode.typing import DfDict, StrList

def long_counts_to_formatted_wide_df_dict(long_df: pd.DataFrame, row_byvar: str, col_byvar: str,
                                          sort_cols_as_numeric: bool =True, sort_rows_as_numeric: bool =True,
                                          sort_cols_as_portvar: bool=False, sort_rows_as_portvar: bool=False,
                                          format_strs: StrList = None
                                          ) -> DfDict:

    if format_strs is None:
        format_strs = ['.0f', '.1f']

    output_cols = [col for col in long_df.columns if col not in [row_byvar, col_byvar]]

    if len(format_strs) != len(output_cols):
        raise ValueError(f'expected {len(output_cols)} format strs, only got {len(format_strs)}: {format_strs}')

    df_dict = {}
    for i, col in enumerate(output_cols):
        df_dict[col] = long_counts_to_formatted_wide_df(
            long_df,
            col,
            row_byvar,
            col_byvar,
            sort_cols_as_numeric=sort_cols_as_numeric,
            sort_rows_as_numeric=sort_rows_as_numeric,
            sort_cols_as_portvar=sort_cols_as_portvar,
            sort_rows_as_portvar=sort_rows_as_portvar,
            format_str=format_strs[i]
        )

    return df_dict


def long_counts_to_formatted_wide_df(long_df: pd.DataFrame, col: str, row_byvar: str, col_byvar: str,
                                     sort_cols_as_numeric: bool =True, sort_rows_as_numeric: bool =True,
                                     sort_cols_as_portvar: bool=False, sort_rows_as_portvar: bool=False,
                                     format_str: str ='.0f'
                                     ) -> pd.DataFrame:

    wide_df = _long_counts_to_wide_df(
        long_df,
        col,
        row_byvar,
        col_byvar
    )

    return sort_and_apply_formatting_to_df(
        wide_df,
        sort_cols_as_numeric=sort_cols_as_numeric,
        sort_rows_as_numeric=sort_rows_as_numeric,
        sort_cols_as_portvar=sort_cols_as_portvar,
        sort_rows_as_portvar=sort_rows_as_portvar,
        format_str=format_str
    )


from typing import Optional, List
import pyexlatex.table as lt

def diff_reg_summary_to_latex_table_and_output(summ, entity_var: str,
                                               caption: str = 'Difference-in-Difference Regressions',
                                               extraname: Optional[str] = None,
                                               below_text: Optional[str] = None,
                                               outfolder: Optional[str] = None,
                                               xvars: Optional[List[str]] = None) -> lt.Table:

    if extraname is not None:
        caption = f'{caption} - {extraname}'

    if below_text is None:
        below_text = f"""
        Difference-in-difference regressions are reported. For this sample, the first and last observations for each
        {entity_var} are selected. The first observation is considered before treatment, while the last observation
        is considered after treatment. 
        """

    if xvars is not None:
        below_text += f'Controls include {_comma_and_join(xvars)}.'

    table = lt.Table.from_list_of_lists_of_dfs(
        [[summ.tables[0]]],
        include_index=True,
        caption=caption,
        below_text=below_text
    )

    if outfolder is not None:
        table.to_pdf_and_move(
            outname=caption,
            outfolder=outfolder
        )

    return table

def _comma_and_join(str_list: List[str]) -> str:
    until_and = ', '.join(str_list[:-1])
    return f'{until_and}, and {str_list[-1]}'